
from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class LAYER(TREElement):
    def __init__(self, value):
        super(LAYER, self).__init__()
        self.add_field('LAYER_ID', 's', 3, value)
        self.add_field('BITRATE', 's', 9, value)


class J2KLRAType(TREElement):
    def __init__(self, value):
        super(J2KLRAType, self).__init__()
        self.add_field('ORIG', 's', 1, value)
        self.add_field('NLEVELS_O', 's', 2, value)
        self.add_field('NBANDS_O', 's', 5, value)
        self.add_field('NLAYERS_O', 'd', 3, value)
        self.add_loop('LAYERs', self.NLAYERS_O, LAYER, value)
        if self.ORIG in ['1', '3', '9']:
            self.add_field('NLEVELS_I', 's', 2, value)
            self.add_field('NBANDS_I', 's', 5, value)
            self.add_field('NLAYERS_I', 's', 3, value)


class J2KLRA(TREExtension):
    _tag_value = 'J2KLRA'
    _data_type = J2KLRAType



"""
The Channel definition for CPHD 0.3.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Union, List

from sarpy.io.phase_history.cphd1_elements.base import DEFAULT_STRICT, FLOAT_FORMAT
from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import IntegerDescriptor, FloatDescriptor, \
    SerializableListDescriptor


class ParametersType(Serializable):
    """
    Channel dependent parameters.
    """

    _fields = (
        'SRP_Index', 'NomTOARateSF', 'FxCtrNom', 'BWSavedNom', 'TOASavedNom',
        'TxAnt_Index', 'RcvAnt_Index', 'TWAnt_Index')
    _required = (
        'SRP_Index', 'NomTOARateSF', 'FxCtrNom', 'BWSavedNom', 'TOASavedNom')
    _numeric_format = {
        'NomTOARateSF': FLOAT_FORMAT, 'FxCtrNom': FLOAT_FORMAT, 'BWSavedNom': FLOAT_FORMAT,
        'TOASavedNom': FLOAT_FORMAT}
    # descriptors
    SRP_Index = IntegerDescriptor(
        'SRP_Index', _required, strict=DEFAULT_STRICT,
        docstring='Index to identify the SRP position function used for the '
                  'channel.')  # type: int
    NomTOARateSF = FloatDescriptor(
        'NomTOARateSF', _required, strict=DEFAULT_STRICT,
        docstring='Scale factor to indicate the fraction of the Doppler spectrum '
                  'that is clear.')  # type: float
    FxCtrNom = FloatDescriptor(
        'FxCtrNom', _required, strict=DEFAULT_STRICT,
        docstring='Nominal center transmit frequency associated with the channel (Hz). '
                  'For DomainType = TOA, FxCtrNom is the center frequency for all '
                  'vectors.')  # type: float
    BWSavedNom = FloatDescriptor(
        'BWSavedNom', _required, strict=DEFAULT_STRICT,
        docstring='Nominal transmit bandwidth associated with the channel (Hz). '
                  'For DomainType = TOA, BWSavedNom is the bandwidth saved for all '
                  'vectors.')  # type: float
    TOASavedNom = FloatDescriptor(
        'TOASavedNom', _required, strict=DEFAULT_STRICT,
        docstring='Nominal span in TOA saved for the channel. For DomainType = FX, '
                  'TOASavedNom is the bandwidth saved for all '
                  'vectors.')  # type: float
    TxAnt_Index = IntegerDescriptor(
        'TxAnt_Index', _required, strict=DEFAULT_STRICT,
        docstring='Indicates the Transmit Antenna pattern for data collected to form '
                  'the CPHD channel.')  # type: Union[None, int]
    RcvAnt_Index = IntegerDescriptor(
        'RcvAnt_Index', _required, strict=DEFAULT_STRICT,
        docstring='Indicates the Receive Antenna pattern for data collected to form '
                  'the CPHD channel.')  # type: Union[None, int]
    TWAnt_Index = IntegerDescriptor(
        'TWAnt_Index', _required, strict=DEFAULT_STRICT,
        docstring='Indicates the T wo-way Antenna pattern for data collected to form '
                  'the CPHD channel.')  # type: Union[None, int]

    def __init__(self, SRP_Index=None, NomTOARateSF=None, FxCtrNom=None, BWSavedNom=None,
                 TOASavedNom=None, TxAnt_Index=None, RcvAnt_Index=None, TWAnt_Index=None,
                 **kwargs):
        """

        Parameters
        ----------
        SRP_Index : int
        NomTOARateSF : float
        FxCtrNom : float
        BWSavedNom : float
        TOASavedNom : float
        TxAnt_Index : None|int
        RcvAnt_Index : None|int
        TWAnt_Index : None|int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SRP_Index = SRP_Index
        self.NomTOARateSF = NomTOARateSF
        self.FxCtrNom = FxCtrNom
        self.BWSavedNom = BWSavedNom
        self.TOASavedNom = TOASavedNom
        self.TxAnt_Index = TxAnt_Index
        self.RcvAnt_Index = RcvAnt_Index
        self.TWAnt_Index = TWAnt_Index
        super(ParametersType, self).__init__(**kwargs)


class ChannelType(Serializable):
    """
    Channel specific parameters for CPHD.
    """

    _fields = ('Parameters', )
    _required = ('Parameters', )
    _collections_tags = {'Parameters': {'array': False, 'child_tag': 'Parameters'}}
    # descriptors
    Parameters = SerializableListDescriptor(
        'Parameters', ParametersType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Channel dependent parameter list.')  # type: List[ParametersType]

    def __init__(self, Parameters=None, **kwargs):
        """

        Parameters
        ----------
        Parameters : List[ParametersType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Parameters = Parameters
        super(ChannelType, self).__init__(**kwargs)





from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class BCKGDAType(TREElement):
    def __init__(self, value):
        super(BCKGDAType, self).__init__()
        self.add_field('BGWIDTH', 's', 8, value)
        self.add_field('BGHEIGHT', 's', 8, value)
        self.add_field('BGRED', 's', 8, value)
        self.add_field('BGGREEN', 's', 8, value)
        self.add_field('BGBLUE', 's', 8, value)
        self.add_field('PIXSIZE', 's', 8, value)
        self.add_field('PIXUNITS', 's', 8, value)


class BCKGDA(TREExtension):
    _tag_value = 'BCKGDA'
    _data_type = BCKGDAType



"""
The Per Vector parameters (PVP) definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Michael Stewart, Valkyrie")

from typing import Union, List

import numpy

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import IntegerDescriptor, SerializableDescriptor, \
    SerializableListDescriptor
from sarpy.io.phase_history.cphd1_elements.PVP import PerVectorParameterI8, \
    PerVectorParameterF8, PerVectorParameterXYZ, UserDefinedPVPType
from sarpy.io.phase_history.cphd1_elements.utils import binary_format_string_to_dtype, homogeneous_dtype
from .base import DEFAULT_STRICT


BYTES_PER_WORD = 8


class PerVectorParameterDCXY(Serializable):
    _fields = ('Offset', 'Size', 'Format')
    _required = ('Offset', )
    # descriptors
    Offset = IntegerDescriptor(
        'Offset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='The offset value.')  # type: int

    def __init__(self, Offset=None, **kwargs):
        """

        Parameters
        ----------
        Offset : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Offset = Offset
        super(PerVectorParameterDCXY, self).__init__(**kwargs)

    @property
    def Size(self):
        """
        int: The size of the vector
        """

        return 2

    @property
    def Format(self):
        """
        str: The format of the vector data, constant value 'DCX=F8;DCY=F8;' here.
        """

        return 'DCX=F8;DCY=F8;'


class PerVectorParameterTxLFM(Serializable):
    _fields = ('Offset', 'Size', 'Format')
    _required = ('Offset', )
    # descriptors
    Offset = IntegerDescriptor(
        'Offset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='The offset value.')  # type: int

    def __init__(self, Offset=None, **kwargs):
        """

        Parameters
        ----------
        Offset : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Offset = Offset
        super(PerVectorParameterTxLFM, self).__init__(**kwargs)

    @property
    def Size(self):
        """
        int: The size of the vector, constant value 3 here.
        """

        return 3

    @property
    def Format(self):
        """
        str: The format of the vector data, constant value 'PhiXC=F8;FxC=F8;FxRate=F8;' here.
        """

        return 'PhiXC=F8;FxC=F8;FxRate=F8;'


class TxAntennaType(Serializable):
    _fields = ('TxACX', 'TxACY', 'TxEB')
    _required = _fields
    # descriptors
    TxACX = SerializableDescriptor(
        'TxACX', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    TxACY = SerializableDescriptor(
        'TxACY', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    TxEB = SerializableDescriptor(
        'TxEB', PerVectorParameterDCXY, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterDCXY

    def __init__(self, TxACX=None, TxACY=None, TxEB=None, **kwargs):
        """

        Parameters
        ----------
        TxACX : PerVectorParameterXYZ
        TxACY : PerVectorParameterXYZ
        TxEB : PerVectorParameterDCXY
        kwargs
        """
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxACX = TxACX
        self.TxACY = TxACY
        self.TxEB = TxEB
        super(TxAntennaType, self).__init__(**kwargs)


class RcvAntennaType(Serializable):
    _fields = ('RcvACX', 'RcvACY', 'RcvEB')
    _required = _fields
    # descriptors
    RcvACX = SerializableDescriptor(
        'RcvACX', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    RcvACY = SerializableDescriptor(
        'RcvACY', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    RcvEB = SerializableDescriptor(
        'RcvEB', PerVectorParameterDCXY, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterDCXY

    def __init__(self, RcvACX=None, RcvACY=None, RcvEB=None, **kwargs):
        """

        Parameters
        ----------
        RcvACX : PerVectorParameterXYZ
        RcvACY : PerVectorParameterXYZ
        RcvEB : PerVectorParameterDCXY
        kwargs
        """
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RcvACX = RcvACX
        self.RcvACY = RcvACY
        self.RcvEB = RcvEB
        super(RcvAntennaType, self).__init__(**kwargs)


class TxPulseType(Serializable):
    _fields = ('TxTime', 'TxPos', 'TxVel', 'FX1', 'FX2', 'TXmt', 'TxLFM', 'TxAntenna')
    _required = ('TxTime', 'TxPos', 'TxVel', 'FX1', 'FX2', 'TXmt')
    # descriptors
    TxTime = SerializableDescriptor(
        'TxTime', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    TxPos = SerializableDescriptor(
        'TxPos', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    TxVel = SerializableDescriptor(
        'TxVel', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    FX1 = SerializableDescriptor(
        'FX1', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    FX2 = SerializableDescriptor(
        'FX2', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    TXmt = SerializableDescriptor(
        'TXmt', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    TxLFM = SerializableDescriptor(
        'TxLFM', PerVectorParameterTxLFM, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, PerVectorParameterTxLFM]
    TxAntenna = SerializableDescriptor(
        'TxAntenna', TxAntennaType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, TxAntennaType]

    def __init__(self, TxTime=None, TxPos=None, TxVel=None,
                 FX1=None, FX2=None, TXmt=None, TxLFM=None,
                 TxAntenna=None, **kwargs):
        """

        Parameters
        ----------
        TxTime : PerVectorParameterF8
        TxPos : PerVectorParameterXYZ
        TxVel : PerVectorParameterXYZ
        FX1 : PerVectorParameterF8
        FX2 : PerVectorParameterF8
        TXmt : PerVectorParameterF8
        TxLFM : None|PerVectorParameterTxLFM
        TxAntenna : None|TxAntennaType
        kwargs
        """
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxTime = TxTime
        self.TxPos = TxPos
        self.TxVel = TxVel
        self.FX1 = FX1
        self.FX2 = FX2
        self.TXmt = TXmt
        self.TxLFM = TxLFM
        self.TxAntenna = TxAntenna
        super(TxPulseType, self).__init__(**kwargs)


class PVPType(Serializable):
    _fields = (
        'RcvTime', 'RcvPos', 'RcvVel', 'RefPhi0', 'RefFreq', 'DFIC0',
        'FICRate', 'FRCV1', 'FRCV2', 'DGRGC', 'SIGNAL', 'AmpSF',
        'RcvAntenna', 'TxPulse', 'AddedPVP')
    _required = (
        'RcvTime', 'RcvPos', 'RcvVel', 'RefPhi0', 'RefFreq', 'DFIC0',
        'FICRate', 'FRCV1', 'FRCV2')
    _collections_tags = {'AddedPVP': {'array': False, 'child_tag': 'AddedPVP'}}
    # descriptors
    RcvTime = SerializableDescriptor(
        'RcvTime', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    RcvPos = SerializableDescriptor(
        'RcvPos', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    RcvVel = SerializableDescriptor(
        'RcvVel', PerVectorParameterXYZ, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterXYZ
    RefPhi0 = SerializableDescriptor(
        'RefPhi0', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    RefFreq = SerializableDescriptor(
        'RefFreq', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    DFIC0 = SerializableDescriptor(
        'DFIC0', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    FICRate = SerializableDescriptor(
        'FICRate', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    FRCV1 = SerializableDescriptor(
        'FRCV1', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    FRCV2 = SerializableDescriptor(
        'FRCV2', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: PerVectorParameterF8
    DGRGC = SerializableDescriptor(
        'DGRGC', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, PerVectorParameterF8]
    SIGNAL = SerializableDescriptor(
        'SIGNAL', PerVectorParameterI8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, PerVectorParameterI8]
    AmpSF = SerializableDescriptor(
        'AmpSF', PerVectorParameterF8, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, PerVectorParameterF8]
    RcvAntenna = SerializableDescriptor(
        'RcvAntenna', RcvAntennaType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, RcvAntennaType]
    TxPulse = SerializableDescriptor(
        'TxPulse', TxPulseType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, TxPulseType]
    AddedPVP = SerializableListDescriptor(
        'AddedPVP', UserDefinedPVPType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, List[UserDefinedPVPType]]

    def __init__(self, RcvTime=None, RcvPos=None, RcvVel=None,
                 RefPhi0=None, RefFreq=None, DFIC0=None,
                 FICRate=None, FRCV1=None, FRCV2=None,
                 DGRGC=None, SIGNAL=None, AmpSF=None,
                 RcvAntenna=None, TxPulse=None,
                 AddedPVP=None, **kwargs):
        """

        Parameters
        ----------
        RcvTime : PerVectorParameterF8
        RcvPos : PerVectorParameterXYZ
        RcvVel : PerVectorParameterXYZ
        RefPhi0 : PerVectorParameterF8
        RefFreq : PerVectorParameterF8
        DFIC0 : PerVectorParameterF8
        FICRate : PerVectorParameterF8
        FRCV1 : PerVectorParameterF8
        FRCV2 : PerVectorParameterF8
        DGRGC : None|PerVectorParameterF8
        SIGNAL : None|PerVectorParameterI8
        AmpSF : None|PerVectorParameterF8
        RcvAntenna : None|RcvAntennaType
        TxPulse : None|TxPulseType
        AddedPVP : None|List[UserDefinedPVPType]
        kwargs
        """
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RcvTime = RcvTime
        self.RcvPos = RcvPos
        self.RcvVel = RcvVel
        self.RefPhi0 = RefPhi0
        self.RefFreq = RefFreq
        self.DFIC0 = DFIC0
        self.FICRate = FICRate
        self.FRCV1 = FRCV1
        self.FRCV2 = FRCV2
        self.DGRGC = DGRGC
        self.SIGNAL = SIGNAL
        self.AmpSF = AmpSF
        self.RcvAntenna = RcvAntenna
        self.TxPulse = TxPulse
        self.AddedPVP = AddedPVP
        super(PVPType, self).__init__(**kwargs)

    def get_size(self):
        """
        Gets the size in bytes of each vector.

        Returns
        -------
        int
        """

        def get_num_words(obj):
            sz = getattr(obj, 'Size')
            if sz is not None:
                return sz
            sz = 0

            # noinspection PyProtectedMember
            for fld in obj._fields:
                fld_val = getattr(obj, fld)
                if fld_val is not None:
                    if fld_val.array:
                        for arr_val in fld_val:
                            sz += get_num_words(arr_val)
                    else:
                        sz += get_num_words(fld_val)
            return sz

        return get_num_words(self) * BYTES_PER_WORD

    def get_offset_size_format(self, field):
        """
        Get the Offset (in bytes), Size (in bytes) for the given field,
        as well as the corresponding struct format string.

        Parameters
        ----------
        field : str
            The desired field name.

        Returns
        -------
        None|(int, int, str)
        """

        def osf_tuple(val_in):
            return val_in.Offset*BYTES_PER_WORD, val_in.Size*BYTES_PER_WORD, homogeneous_dtype(val_in.Format).char

        # noinspection PyProtectedMember
        if field in self._fields[:-1]:
            val = getattr(self, field)
            if val is None:
                return None
            return osf_tuple(val)
        elif self.RcvAntenna and field in self.RcvAntenna._fields:
            val = getattr(self.RcvAntenna, field)
            if val is None:
                return None
            return osf_tuple(val)
        elif self.TxPulse and field in self.TxPulse._fields:
            val = getattr(self.TxPulse, field)
            if val is None:
                return None
            return osf_tuple(val)
        elif self.TxPulse and self.TxPulse.TxAntenna and field in self.TxPulse.TxAntenna._fields:
            val = getattr(self.TxPulse.TxAntenna, field)
            if val is None:
                return None
            return osf_tuple(val)
        else:
            if self.AddedPVP is None:
                return None
            for val in self.AddedPVP:
                if field == val.Name:
                    return osf_tuple(val)
        return None

    def get_vector_dtype(self):
        """
        Gets the dtype for the corresponding structured array for the full PVP set.

        Returns
        -------
        numpy.dtype
            This will be a compound dtype for a structured array.
        """

        names = []
        formats = []
        offsets = []

        for field in self._fields:
            val = getattr(self, field)
            if val is None:
                continue
            elif field == "AddedPVP":
                for entry in val:
                    names.append(entry.Name)
                    formats.append(binary_format_string_to_dtype(entry.Format))
                    offsets.append(entry.Offset*BYTES_PER_WORD)
            elif field == 'RcvAntenna' or field == 'TxPulse':
                continue
            else:
                names.append(field)
                formats.append(binary_format_string_to_dtype(val.Format))
                offsets.append(val.Offset*BYTES_PER_WORD)

        if self.RcvAntenna is not None:
            # noinspection PyProtectedMember
            for field in self.RcvAntenna._fields:
                val = getattr(self.RcvAntenna, field)
                if val is None:
                    continue
                else:
                    names.append(field)
                    formats.append(binary_format_string_to_dtype(val.Format))
                    offsets.append(val.Offset*BYTES_PER_WORD)

        if self.TxPulse is not None:
            # noinspection PyProtectedMember
            for field in self.TxPulse._fields:
                val = getattr(self.TxPulse, field)
                if val is None:
                    continue
                elif field == 'TxAntenna':
                    continue
                else:
                    names.append(field)
                    formats.append(binary_format_string_to_dtype(val.Format))
                    offsets.append(val.Offset*BYTES_PER_WORD)
            if self.TxPulse.TxAntenna is not None:
                # noinspection PyProtectedMember
                for field in self.TxPulse.TxAntenna._fields:
                    val = getattr(self.TxPulse.TxAntenna, field)
                    if val is None:
                        continue
                    else:
                        names.append(field)
                        formats.append(binary_format_string_to_dtype(val.Format))
                        offsets.append(val.Offset*BYTES_PER_WORD)

        return numpy.dtype({'names': names, 'formats': formats, 'offsets': offsets})




__classification__ = "UNCLASSIFIED"


def open(file_name: str):
    """
    Given a file, try to find and return the appropriate reader object.

    Parameters
    ----------
    file_name : str

    Returns
    -------
    BaseReader

    Raises
    ------
    SarpyIOError
    """

    from .complex.converter import open_complex
    from .product.converter import open_product
    from .phase_history.converter import open_phase_history
    from .received.converter import open_received
    from .general.converter import open_general
    from .general.base import SarpyIOError

    try:
        return open_complex(file_name)
    except SarpyIOError:
        pass

    try:
        return open_product(file_name)
    except SarpyIOError:
        pass

    try:
        return open_phase_history(file_name)
    except SarpyIOError:
        pass

    try:
        return open_received(file_name)
    except SarpyIOError:
        pass

    try:
        return open_general(file_name)
    except SarpyIOError:
        pass

    raise SarpyIOError(
        'The format of file {} does not match any reader in the complex, '
        'product, phase_history, received, or general modules.'.format(file_name))



"""
The error parameters type definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Michael Stewart, Valkyrie")

from typing import Union

from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import FloatDescriptor, SerializableDescriptor, \
    ParametersDescriptor

from sarpy.io.complex.sicd_elements.blocks import ErrorDecorrFuncType
from sarpy.io.complex.sicd_elements.ErrorStatistics import PosVelErrType, TropoErrorType

from .base import DEFAULT_STRICT, FLOAT_FORMAT


class RadarSensorType(Serializable):
    """
    Radar sensor error statistics.
    """

    _fields = ('RangeBias', 'ClockFreqSF', 'RefTimeError', 'RangeBiasDecorr')
    _required = ('RangeBias', )
    _numeric_format = {'RangeBias': FLOAT_FORMAT, 'ClockFreqSF': FLOAT_FORMAT, 'RefTimeError': FLOAT_FORMAT}
    # descriptors
    RangeBias = FloatDescriptor(
        'RangeBias', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Range bias error standard deviation.')  # type: float
    ClockFreqSF = FloatDescriptor(
        'ClockFreqSF', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Payload clock frequency scale factor standard deviation, '
                  r'where :math:`SF = (\Delta f)/f_0`.')  # type: float
    RefTimeError = FloatDescriptor(
        'RefTimeError', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Time error standard deviation, in seconds.')  # type: float
    RangeBiasDecorr = SerializableDescriptor(
        'RangeBiasDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Range Bias error decorrelation function.')  # type: ErrorDecorrFuncType

    def __init__(self, RangeBias=None, ClockFreqSF=None, RefTimeError=None,
                 RangeBiasDecorr=None, **kwargs):
        """

        Parameters
        ----------
        RangeBias : float
        ClockFreqSF : None|float
        RefTimeError : None|float
        RangeBiasDecorr : None|ErrorDecorrFuncType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RangeBias = RangeBias
        self.ClockFreqSF = ClockFreqSF
        self.RefTimeError = RefTimeError
        self.RangeBiasDecorr = RangeBiasDecorr
        super(RadarSensorType, self).__init__(**kwargs)


class IonoErrorType(Serializable):
    """
    Ionosphere delay error statistics.
    """

    _fields = ('IonoRangeVertical', 'IonoRangeRateVertical', 'IonoRgRgRateCC', 'IonoRangeVertDecorr')
    _required = ('IonoRangeVertical', )
    _numeric_format = {'IonoRangeVertical': FLOAT_FORMAT, 'IonoRangeRateVertical': FLOAT_FORMAT, 'IonoRgRgRateCC': FLOAT_FORMAT}
    # descriptors
    IonoRangeVertical = FloatDescriptor(
        'IonoRangeVertical', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Ionosphere two-way delay error for normal incidence standard deviation. '
                  r'Expressed as a range error. :math:`(\Delta R) = (\Delta T) \cdot (c/2)`.')  # type: float
    IonoRangeRateVertical = FloatDescriptor(
        'IonoRangeRateVertical', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Ionosphere two-way delay rate of change error for normal incidence standard deviation. '
                  r'Expressed as a range rate error. :math:`\dot{R} = \Delta \dot{TD_Iono} \times c/2`.')  # type: float
    IonoRgRgRateCC = FloatDescriptor(
        'IonoRgRgRateCC', _required, strict=DEFAULT_STRICT, bounds=(-1, 1),
        docstring='Ionosphere range error and range rate error correlation coefficient.')  # type: float
    IonoRangeVertDecorr = SerializableDescriptor(
        'IonoRangeVertDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere range error decorrelation function.')  # type: ErrorDecorrFuncType

    def __init__(self, IonoRangeVertical=None, IonoRangeRateVertical=None,
                 IonoRgRgRateCC=None, IonoRangeVertDecorr=None, **kwargs):
        """

        Parameters
        ----------
        IonoRangeVertical : float
        IonoRangeRateVertical : None|float
        IonoRgRgRateCC : None|float
        IonoRangeVertDecorr : None|ErrorDecorrFuncType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.IonoRangeVertical = IonoRangeVertical
        self.IonoRangeRateVertical = IonoRangeRateVertical
        self.IonoRgRgRateCC = IonoRgRgRateCC
        self.IonoRangeVertDecorr = IonoRangeVertDecorr
        super(IonoErrorType, self).__init__(**kwargs)


class BistaticRadarSensorType(Serializable):
    """
    Error statistics for a single radar platform.
    """

    _fields = ('ClockFreqSF', 'RefTimeError')
    _required = tuple()
    _numeric_format = {'ClockFreqSF': FLOAT_FORMAT, 'RefTimeError': FLOAT_FORMAT}
    # descriptors
    ClockFreqSF = FloatDescriptor(
        'ClockFreqSF', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Payload clock frequency scale factor standard deviation, '
                  r'where :math:`SF = (\Delta f)/f_0`.')  # type: float
    RefTimeError = FloatDescriptor(
        'RefTimeError', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Time error standard deviation, in seconds.')  # type: float

    def __init__(self, ClockFreqSF=None, RefTimeError=None, **kwargs):
        """

        Parameters
        ----------
        ClockFreqSF : None|float
        RefTimeError : None|float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ClockFreqSF = ClockFreqSF
        self.RefTimeError = RefTimeError
        super(BistaticRadarSensorType, self).__init__(**kwargs)


class ReceiveSensorType(Serializable):
    """
    Receive sensor error statistics
    """

    _fields = ('RcvDelayBias', 'ClockFreqSF', 'RefTimeError')
    _required = ('RcvDelayBias', )
    _numeric_format = {'ClockFreqSF': FLOAT_FORMAT, 'RefTimeError': FLOAT_FORMAT}
    # descriptors
    RcvDelayBias = FloatDescriptor(
        'RcvDelayBias', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Receive path signal delay error bias that causes an error in the'
                  ' estimated signal time of arrival at the Receive APC')  # type: float
    ClockFreqSF = FloatDescriptor(
        'ClockFreqSF', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Payload clock frequency scale factor standard deviation, '
                  r'where :math:`SF = (\Delta f)/f_0`.')  # type: float
    RefTimeError = FloatDescriptor(
        'RefTimeError', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Time error standard deviation, in seconds.')  # type: float

    def __init__(self, RcvDelayBias=None, ClockFreqSF=None, RefTimeError=None, **kwargs):
        """

        Parameters
        ----------
        RcvDelayBias : float
        ClockFreqSF : None|float
        RefTimeError : None|float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RcvDelayBias = RcvDelayBias
        self.ClockFreqSF = ClockFreqSF
        self.RefTimeError = RefTimeError
        super(ReceiveSensorType, self).__init__(**kwargs)


class MonostaticType(Serializable):
    """
    Error parameters for monostatic collection.
    """

    _fields = ('PosVelErr', 'RadarSensor', 'TropoError', 'IonoError', 'AddedParameters')
    _required = ('PosVelErr', 'RadarSensor')
    _collections_tags = {'AddedParameters': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    PosVelErr = SerializableDescriptor(
        'PosVelErr', PosVelErrType, _required, strict=DEFAULT_STRICT,
        docstring='Position and velocity error statistics for the sensor '
                  'platform.')  # type: PosVelErrType
    RadarSensor = SerializableDescriptor(
        'RadarSensor', RadarSensorType, _required, strict=DEFAULT_STRICT,
        docstring='Radar sensor error statistics.')  # type: RadarSensorType
    TropoError = SerializableDescriptor(
        'TropoError', TropoErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Troposphere delay error statistics.')  # type: TropoErrorType
    IonoError = SerializableDescriptor(
        'IonoError', IonoErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere delay error statistics.')  # type: IonoErrorType
    AddedParameters = ParametersDescriptor(
        'AddedParameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Additional error parameters.')  # type: ParametersCollection

    def __init__(self, PosVelErr=None, RadarSensor=None, TropoError=None, IonoError=None,
                 AddedParameters=None, **kwargs):
        """

        Parameters
        ----------
        PosVelErr : PosVelErrType
        RadarSensor : RadarSensorType
        TropoError : None|TropoErrorType
        IonoError : None|IonoErrorType
        AddedParameters : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PosVelErr = PosVelErr
        self.RadarSensor = RadarSensor
        self.TropoError = TropoError
        self.IonoError = IonoError
        self.AddedParameters = AddedParameters
        super(MonostaticType, self).__init__(PosVelErr=PosVelErr, RadarSensor=RadarSensor, **kwargs)


class PlatformType(Serializable):
    """
    Basic bistatic platform error type definition.
    """
    _fields = ('PosVelErr', 'RadarSensor')
    _required = _fields
    # descriptors
    PosVelErr = SerializableDescriptor(
        'PosVelErr', PosVelErrType, _required, strict=DEFAULT_STRICT,
        docstring='Position and velocity error statistics for the sensor '
                  'platform.')  # type: PosVelErrType
    RadarSensor = SerializableDescriptor(
        'RadarSensor', BistaticRadarSensorType, _required, strict=DEFAULT_STRICT,
        docstring='Platform sensor error statistics.')  # type: BistaticRadarSensorType

    def __init__(self, PosVelErr=None, RadarSensor=None, **kwargs):
        """

        Parameters
        ----------
        PosVelErr : PosVelErrType
        RadarSensor : BistaticRadarSensorType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PosVelErr = PosVelErr
        self.RadarSensor = RadarSensor
        super(PlatformType, self).__init__(**kwargs)


class BistaticType(Serializable):
    """
    Error parameters for bistatic parameters.
    """

    _fields = ('TxPlatform', 'RcvPlatform', 'AddedParameters')
    _required = ('TxPlatform', 'RcvPlatform')
    _collections_tags = {'AddedParameters': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    TxPlatform = SerializableDescriptor(
        'TxPlatform', PlatformType, _required, strict=DEFAULT_STRICT,
        docstring='Error statistics for the transmit platform.')  # type: PlatformType
    RcvPlatform = SerializableDescriptor(
        'RcvPlatform', PlatformType, _required, strict=DEFAULT_STRICT,
        docstring='Error statistics for the receive platform.')  # type: PlatformType
    AddedParameters = ParametersDescriptor(
        'AddedParameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Additional error parameters.')  # type: ParametersCollection

    def __init__(self, TxPlatform=None, RcvPlatform=None, AddedParameters=None, **kwargs):
        """

        Parameters
        ----------
        TxPlatform : PlatformType
        RcvPlatform : PlatformType
        AddedParameters : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxPlatform = TxPlatform
        self.RcvPlatform = RcvPlatform
        self.AddedParameters = AddedParameters
        super(BistaticType, self).__init__(**kwargs)


class ReceiveOnlyType(Serializable):
    """
    Parameters for receive only collect type
    """

    _fields = ('PosVelErr', 'ReceiveSensorType', 'AddedParameters')
    _required = ('PosVelErr', 'ReceiveSensorType')
    _collections_tags = {'AddedParameters': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    PosVelErr = SerializableDescriptor(
        'PosVelErr', PosVelErrType, _required, strict=DEFAULT_STRICT,
        docstring='Position and velocity error statistics for the sensor '
                  'platform.')  # type: PosVelErrType
    ReceiveSensor = SerializableDescriptor(
        'ReceiveSensor', RadarSensorType, _required, strict=DEFAULT_STRICT,
        docstring='Receive sensor error statistics.')  # type: RadarSensorType
    AddedParameters = ParametersDescriptor(
        'AddedParameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Additional error parameters.')  # type: ParametersCollection

    def __init__(self, PosVelErr=None, ReceiveSensor=None, AddedParameters=None, **kwargs):
        """

        Parameters
        ----------
        PosVelErr : PosVelErrType
        ReceiveSensor : RadarSensorType
        AddedParameters : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PosVelErr = PosVelErr
        self.ReceiveSensor = ReceiveSensor
        self.AddedParameters = AddedParameters
        super(ReceiveOnlyType, self).__init__(**kwargs)


class ErrorParametersType(Serializable):
    """
    Parameters that describe the statistics of errors in measured or estimated
    parameters that describe the collection.
    """

    _fields = ('Monostatic', 'Bistatic', 'ReceiveOnly')
    _required = ()
    _choice = ({'required': True, 'collection': _fields}, )

    # descriptors
    Monostatic = SerializableDescriptor(
        'Monostatic', MonostaticType, _required, strict=DEFAULT_STRICT,
        docstring='The monostatic parameters.')  # type: Union[None, MonostaticType]
    Bistatic = SerializableDescriptor(
        'Bistatic', BistaticType, _required, strict=DEFAULT_STRICT,
        docstring='The bistatic parameters.')  # type: Union[None, BistaticType]
    ReceiveOnly = SerializableDescriptor(
        'ReceiveOnly', ReceiveOnlyType, _required, strict=DEFAULT_STRICT,
        docstring='The receive only parameters.')  # type: Union[None, BistaticType]

    def __init__(self, Monostatic=None, Bistatic=None, ReceiveOnly=None, **kwargs):
        """

        Parameters
        ----------
        Monostatic : None|MonostaticType
        Bistatic : None|BistaticType
        ReceiveOnly : None|ReceiveOnlyType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Monostatic = Monostatic
        self.Bistatic = Bistatic
        self.ReceiveOnly = ReceiveOnly
        super(ErrorParametersType, self).__init__(**kwargs)



"""
This package contains the elements for interpreting phase history data.
"""

__classification__ = 'UNCLASSIFIED'


def open(*args, **kwargs):
    from .converter import open_phase_history
    return open_phase_history(*args, **kwargs)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class AIMIDA_69Type(TREElement):
    def __init__(self, value):
        super(AIMIDA_69Type, self).__init__()
        self.add_field('MISSION_DATE', 's', 7, value)
        self.add_field('MISSION_NO', 's', 4, value)
        self.add_field('FLIGHT_NO', 's', 2, value)
        self.add_field('OP_NUM', 's', 3, value)
        self.add_field('START_SEGMENT', 's', 2, value)
        self.add_field('REPRO_NUM', 's', 2, value)
        self.add_field('REPLAY', 's', 3, value)
        self.add_field('RESVD001', 's', 1, value)
        self.add_field('START_COLUMN', 's', 2, value)
        self.add_field('START_ROW', 's', 5, value)
        self.add_field('END_SEGMENT', 's', 2, value)
        self.add_field('END_COLUMN', 's', 2, value)
        self.add_field('END_ROW', 's', 5, value)
        self.add_field('COUNTRY', 's', 2, value)
        self.add_field('RESVD002', 's', 4, value)
        self.add_field('LOCATION', 's', 11, value)
        self.add_field('TIME', 's', 5, value)
        self.add_field('CREATION_DATE', 's', 7, value)


class AIMIDA_69(TREExtension):
    _tag_value = 'AIMIDA'
    _data_type = AIMIDA_69Type


class AIMIDA_73Type(TREElement):
    def __init__(self, value):
        super(AIMIDA_73Type, self).__init__()
        self.add_field('MISSION_DATE', 's', 8, value)
        self.add_field('MISSION_NO', 's', 4, value)
        self.add_field('FLIGHT_NO', 's', 2, value)
        self.add_field('OP_NUM', 's', 3, value)
        self.add_field('RESVD001', 's', 2, value)
        self.add_field('REPRO_NUM', 's', 2, value)
        self.add_field('REPLAY', 's', 3, value)
        self.add_field('RESVD002', 's', 1, value)
        self.add_field('START_COLUMN', 's', 3, value)
        self.add_field('START_ROW', 's', 5, value)
        self.add_field('RESVD003', 's', 2, value)
        self.add_field('END_COLUMN', 's', 3, value)
        self.add_field('END_ROW', 's', 5, value)
        self.add_field('COUNTRY', 's', 2, value)
        self.add_field('RESVD004', 's', 4, value)
        self.add_field('LOCATION', 's', 11, value)
        self.add_field('TIME', 's', 5, value)
        self.add_field('CREATION_DATE', 's', 8, value)


class AIMIDA_73(TREExtension):
    _tag_value = 'AIMIDA'
    _data_type = AIMIDA_73Type


class AIMIDA_89Type(TREElement):
    def __init__(self, value):
        super(AIMIDA_89Type, self).__init__()
        self.add_field('MISSION_DATE_TIME', 's', 14, value)
        self.add_field('MISSION_NO', 's', 4, value)
        self.add_field('MISSION_ID', 's', 10, value)
        self.add_field('FLIGHT_NO', 's', 2, value)
        self.add_field('OP_NUM', 's', 3, value)
        self.add_field('CURRENT_SEGMENT', 's', 2, value)
        self.add_field('REPRO_NUM', 's', 2, value)
        self.add_field('REPLAY', 's', 3, value)
        self.add_field('RESVD001', 's', 1, value)
        self.add_field('START_COLUMN', 's', 3, value)
        self.add_field('START_ROW', 's', 5, value)
        self.add_field('END_SEGMENT', 's', 2, value)
        self.add_field('END_COLUMN', 's', 3, value)
        self.add_field('END_ROW', 's', 5, value)
        self.add_field('COUNTRY', 's', 2, value)
        self.add_field('RESVD002', 's', 4, value)
        self.add_field('LOCATION', 's', 11, value)
        self.add_field('RESVD003', 's', 13, value)


class AIMIDA_89(TREExtension):
    _tag_value = 'AIMIDA'
    _data_type = AIMIDA_89Type


class AIMIDA(TREExtension):
    _tag_value = 'AIMIDA'

    def __init__(self):
        raise ValueError(
            'Not to be implemented directly. '
            'Use of one AIMIDA_69, AIMIDA_73, or AIMIDA_89')

    @classmethod
    def from_bytes(cls, value, start):
        """

        Parameters
        ----------
        value : bytes
        start : int

        Returns
        -------
        AIMIDA_69|AIMIDA_73|AIMIDA_89
        """

        tag_value = value[start:start+6].decode('utf-8').strip()
        if tag_value != cls._tag_value:
            raise ValueError('tag value must be {}. Got {}'.format(cls._tag_value, tag_value))

        lng = int(value[start+6:start+11])
        if lng == 69:
            return AIMIDA_69.from_bytes(value, start)
        elif lng == 73:
            return AIMIDA_73.from_bytes(value, start)
        elif lng == 89:
            return AIMIDA_89.from_bytes(value, start)
        else:
            raise ValueError('the data must be length 69, 73, or 89. Got {}'.format(lng))



"""
The module contains methods for computing a coherent change detection from registered images
"""

from typing import Union, Tuple

import numpy
import scipy.signal

__classification__ = "UNCLASSIFIED"
__author__ = ('Thomas Mccullough',  'Wade Schwartzkopf', 'Mike Dowell')


def mem(
        reference_image: numpy.ndarray,
        match_image: numpy.ndarray,
        corr_window_size: Union[int, Tuple[int, int]]) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Performs coherent change detection, following the equation as described in
    Jakowatz, et al., "Spotlight-mode Synthetic Aperture radar: A Signal
    Processing Approach".

    .. warning: This assumes that the two arrays have already been properly
        registered with respect to one another, and all processing will proceed
        directly in memory.

    Parameters
    ----------
    reference_image : numpy.ndarray
    match_image : numpy.ndarray
    corr_window_size : int|tuple
        The correlation window size. If int, a square correlation window of
        given size will be used. If tuple, it must be a two element tuple of
        ints which describe the correlation window size.

    Returns
    -------
    (numpy.ndarray, numpy.ndarray)
        The ccd and phase arrays
    """

    if isinstance(corr_window_size, int):
        kernel = numpy.ones((corr_window_size, corr_window_size), dtype=numpy.float32)
    elif isinstance(corr_window_size, tuple) and len(corr_window_size) == 2:
        kernel = numpy.ones(corr_window_size, dtype=numpy.float32)
    else:
        raise TypeError('corr_window_size is required to be an int or two element tuple of ints')

    inner_product = scipy.signal.convolve2d(
        numpy.conj(reference_image)*match_image, kernel, mode='same')
    # calculate magnitude of smeared reference image, accounting for numerical errors
    ref_mag = scipy.signal.convolve2d(
        reference_image.real*reference_image.real + reference_image.imag*reference_image.imag,
        kernel, mode='same')
    ref_mag[ref_mag < 0] = 0
    ref_mag = numpy.sqrt(ref_mag)
    # same for match image
    match_mag = scipy.signal.convolve2d(
        match_image.real*match_image.real + match_image.imag*match_image.imag,
        kernel, mode='same')
    match_mag[match_mag < 0] = 0
    match_mag = numpy.sqrt(match_mag)
    # perform the ccd calculation
    ccd = numpy.where((ref_mag > 0) & (match_mag > 0), inner_product/(ref_mag*match_mag), numpy.float32(0.0))
    phase = numpy.angle(inner_product)
    return ccd, phase




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class CSEXRAType(TREElement):
    def __init__(self, value):
        super(CSEXRAType, self).__init__()
        self.add_field('SENSOR', 's', 6, value)
        self.add_field('TIME_FIRST_LINE_IMAGE', 's', 12, value)
        self.add_field('TIME_IMAGE_DURATION', 's', 12, value)
        self.add_field('MAX_GSD', 's', 5, value)
        self.add_field('ALONG_SCAN_GSD', 's', 5, value)
        self.add_field('CROSS_SCAN_GSD', 's', 5, value)
        self.add_field('GEO_MEAN_GSD', 's', 5, value)
        self.add_field('A_S_VERT_GSD', 's', 5, value)
        self.add_field('C_S_VERT_GSD', 's', 5, value)
        self.add_field('GEO_MEAN_VERT_GSD', 's', 5, value)
        self.add_field('GSD_BETA_ANGLE', 's', 5, value)
        self.add_field('DYNAMIC_RANGE', 's', 5, value)
        self.add_field('NUM_LINES', 's', 7, value)
        self.add_field('NUM_SAMPLES', 's', 5, value)
        self.add_field('ANGLE_TO_NORTH', 's', 7, value)
        self.add_field('OBLIQUITY_ANGLE', 's', 6, value)
        self.add_field('AZ_OF_OBLIQUITY', 's', 7, value)
        self.add_field('GRD_COVER', 's', 1, value)
        self.add_field('SNOW_DEPTH_CAT', 's', 1, value)
        self.add_field('SUN_AZIMUTH', 's', 7, value)
        self.add_field('SUN_ELEVATION', 's', 7, value)
        self.add_field('PREDICTED_NIIRS', 's', 3, value)
        self.add_field('CIRCL_ERR', 's', 3, value)
        self.add_field('LINEAR_ERR', 's', 3, value)


class CSEXRA(TREExtension):
    _tag_value = 'CSEXRA'
    _data_type = CSEXRAType



"""
The GeoDataType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from collections import OrderedDict
from typing import Union, List
from xml.etree import ElementTree

from sarpy.io.xml.base import Serializable, SerializableArray, find_children
from sarpy.io.xml.descriptors import SerializableArrayDescriptor, StringEnumDescriptor
from sarpy.io.complex.sicd_elements.base import SerializableCPArray, SerializableCPArrayDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import LatLonCornerStringType, LatLonArrayElementType, GeoInfoType


class GeoDataType(Serializable):
    """
    Container specifying the image coverage area in geographic coordinates.

    .. Note: The SICD.GeoData class is an extension of this class. Implementation remain
        separate to allow the possibility of different functionality.
    """

    _fields = ('EarthModel', 'ImageCorners', 'ValidData')
    _required = ('EarthModel', 'ImageCorners', 'ValidData')
    _collections_tags = {
        'ValidData': {'array': True, 'child_tag': 'Vertex'},
        'ImageCorners': {'array': True, 'child_tag': 'ICP'}}
    _numeric_format = {'ImageCorners': FLOAT_FORMAT, 'ValidData': FLOAT_FORMAT}
    _child_xml_ns_key = {'GeoInfos': 'sicommon'}
    # other class variables
    _EARTH_MODEL_VALUES = ('WGS_84', )
    # descriptors
    EarthModel = StringEnumDescriptor(
        'EarthModel', _EARTH_MODEL_VALUES, _required, strict=True, default_value='WGS_84',
        docstring='Identifies the earth model used for latitude, longitude and height parameters. '
                  'All height values are *Height Above The Ellipsoid '
                  '(HAE)*.'.format(_EARTH_MODEL_VALUES))  # type: str
    ImageCorners = SerializableCPArrayDescriptor(
        'ImageCorners', LatLonCornerStringType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The geographic image corner points array. Image corners points projected to the '
                  'ground/surface level. Points may be projected to the same height as the SCP if ground/surface '
                  'height data is not available. The corner positions are approximate geographic locations and '
                  'not intended for analytical '
                  'use.')  # type: Union[SerializableCPArray, List[LatLonCornerStringType]]
    ValidData = SerializableArrayDescriptor(
        'ValidData', LatLonArrayElementType, _collections_tags, _required,
        strict=DEFAULT_STRICT, minimum_length=3,
        docstring='The full image array includes both valid data and some zero filled pixels. Simple convex '
                  'polygon enclosed the valid data (may include some zero filled pixels for simplification). '
                  'Vertices in clockwise order.')  # type: Union[SerializableArray, List[LatLonArrayElementType]]

    def __init__(self, EarthModel='WGS_84', ImageCorners=None, ValidData=None, GeoInfos=None, **kwargs):
        """

        Parameters
        ----------
        EarthModel : str
        ImageCorners : SerializableCPArray|List[LatLonCornerStringType]|numpy.ndarray|list|tuple
        ValidData : SerializableArray|List[LatLonArrayElementType]|numpy.ndarray|list|tuple
        GeoInfos : List[GeoInfoType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.EarthModel = EarthModel
        self.ImageCorners = ImageCorners
        self.ValidData = ValidData

        self._GeoInfos = []
        if GeoInfos is None:
            pass
        elif isinstance(GeoInfos, GeoInfoType):
            self.setGeoInfo(GeoInfos)
        elif isinstance(GeoInfos, (list, tuple)):
            for el in GeoInfos:
                self.setGeoInfo(el)
        else:
            raise ValueError('GeoInfos got unexpected type {}'.format(type(GeoInfos)))
        super(GeoDataType, self).__init__(**kwargs)

    @property
    def GeoInfos(self):
        """
        List[GeoInfoType]: list of GeoInfos.
        """

        return self._GeoInfos

    def getGeoInfo(self, key):
        """
        Get the GeoInfo(s) with name attribute == `key`

        Parameters
        ----------
        key : str

        Returns
        -------
        List[GeoInfoType]
        """

        return [entry for entry in self._GeoInfos if entry.name == key]

    def setGeoInfo(self, value):
        """
        Add the given GeoInfo to the GeoInfos list.

        Parameters
        ----------
        value : GeoInfoType

        Returns
        -------
        None
        """
        if isinstance(value, ElementTree.Element):
            gi_key = self._child_xml_ns_key.get('GeoInfos', self._xml_ns_key)
            value = GeoInfoType.from_node(value, self._xml_ns, ns_key=gi_key)
        elif isinstance(value, dict):
            value = GeoInfoType.from_dict(value)

        if isinstance(value, GeoInfoType):
            self._GeoInfos.append(value)
        else:
            raise TypeError('Trying to set GeoInfo element with unexpected type {}'.format(type(value)))

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = OrderedDict()
        gkey = cls._child_xml_ns_key.get('GeoInfos', ns_key)
        kwargs['GeoInfos'] = find_children(node, 'GeoInfo', xml_ns, gkey)
        return super(GeoDataType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        node = super(GeoDataType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity, strict=strict, exclude=exclude)
        # slap on the GeoInfo children
        gkey = self._child_xml_ns_key.get('GeoInfos', ns_key)
        for entry in self._GeoInfos:
            entry.to_node(doc, 'GeoInfo', ns_key=gkey, parent=node, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(GeoDataType, self).to_dict(
            check_validity=check_validity, strict=strict, exclude=exclude)
        # slap on the GeoInfo children
        if len(self.GeoInfos) > 0:
            out['GeoInfos'] = [entry.to_dict(check_validity=check_validity, strict=strict) for entry in self._GeoInfos]
        return out



"""
Definition for the SensorInfo NGA modified RDE/AFRL labeling object
"""

__classification__ = "UNCLASSIFIED"
__authors__ = "Thomas McCullough"


from typing import Optional
import numpy

from sarpy.io.xml.base import Serializable, Arrayable
from sarpy.io.xml.descriptors import SerializableDescriptor, StringDescriptor, \
    StringEnumDescriptor, FloatDescriptor
from sarpy.io.complex.sicd_elements.blocks import XYZType
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.geometry.geocoords import ecf_to_geodetic, ecf_to_ned

from .blocks import LatLonEleType


class BeamWidthType(Serializable, Arrayable):
    _fields = ('Azimuth', 'Elevation')
    _required = _fields
    _numeric_format = {key: '0.17G' for key in _fields}
    # Descriptors
    Azimuth = FloatDescriptor(
        'Azimuth', _required, strict=True, docstring='The Azimuth attribute.')  # type: float
    Elevation = FloatDescriptor(
        'Elevation', _required, strict=True, docstring='The Elevation attribute.')  # type: float

    def __init__(self, Azimuth=None, Elevation=None, **kwargs):
        """
        Parameters
        ----------
        Azimuth : float
        Elevation : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Azimuth, self.Elevation = Azimuth, Elevation
        super(BeamWidthType, self).__init__(**kwargs)

    def get_array(self, dtype='float64'):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [Azimuth, Elevation]
        """

        return numpy.array([self.Azimuth, self.Elevation], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Azimuth, Elevation]

        Returns
        -------
        BeamWidthType
        """
        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 2:
                raise ValueError('Expected array to be of length 2, and received {}'.format(array))
            return cls(Azimuth=array[0], Elevation=array[1])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class SquintAngleType(Serializable):
    _fields = ('GroundPlane', 'SlantPlane')
    _required = _fields
    _numeric_format = {el: '0.17G' for el in _fields}
    # descriptor
    GroundPlane = FloatDescriptor(
        'GroundPlane', _required,
        docstring='Measured angle between the sensor line-of-sight and the '
                  'lateral axis of the aircraft as projected into the'
                  'ground plane')  # type: float
    SlantPlane = FloatDescriptor(
        'SlantPlane', _required,
        docstring='Measured angle between the sensor line-of-sight and the '
                  'lateral axis of the aircraft as projected into the'
                  'slant plane')  # type: float

    def __init__(self, GroundPlane=None, SlantPlane=None, **kwargs):
        """
        Parameters
        ----------
        GroundPlane : float
        SlantPlane : float
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.GroundPlane = GroundPlane
        self.SlantPlane = SlantPlane
        super(SquintAngleType, self).__init__(**kwargs)


class AircraftLocationType(Serializable, Arrayable):
    """A three-dimensional geographic point in WGS-84 coordinates."""
    _fields = ('Lat', 'Lon', 'Altitude')
    _required = _fields
    _numeric_format = {'Lat': '0.17G', 'Lon': '0.17G', 'Altitude': '0.17G'}
    # descriptors
    Lat = FloatDescriptor(
        'Lat', _required, strict=True,
        docstring='The latitude attribute. Assumed to be WGS-84 coordinates.'
    )  # type: float
    Lon = FloatDescriptor(
        'Lon', _required, strict=True,
        docstring='The longitude attribute. Assumed to be WGS-84 coordinates.'
    )  # type: float
    Altitude = FloatDescriptor(
        'Altitude', _required, strict=True,
        docstring='The Height Above Ellipsoid (in meters) attribute. '
                  'Assumed to be WGS-84 coordinates.')  # type: float

    def __init__(self, Lat=None, Lon=None, Altitude=None, **kwargs):
        """
        Parameters
        ----------
        Lat : float
        Lon : float
        Altitude : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Lat = Lat
        self.Lon = Lon
        self.Altitude = Altitude
        super(AircraftLocationType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64):
        """
        Gets an array representation of the data.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            data type of the return

        Returns
        -------
        numpy.ndarray
            data array with appropriate entry order
        """

        return numpy.array([self.Lat, self.Lon, self.Altitude], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Lat, Lon, Altitude]

        Returns
        -------
        AircraftLocationType
        """
        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) < 3:
                raise ValueError('Expected array to be of length 3, and received {}'.format(array))
            return cls(Lat=array[0], Lon=array[1], Altitude=array[2])
        raise ValueError('Expected array to be numpy.ndarray, list, or tuple, got {}'.format(type(array)))


class SensorInfoType(Serializable):
    _fields = (
        'Name', 'SensorMfg', 'OperatingAgency', 'Type', 'Mode', 'Band',
        'Bandwidth', 'CenterFrequency', 'NearRange', 'SlantRangeSwathWidth',
        'Polarization', 'Range', 'DepressionAngle', 'LinearDynamicRange',
        'BeamWidth', 'Aimpoint', 'AircraftHeading', 'AircraftTrackAngle', 'Look', 'SquintAngle',
        'AircraftLocation', 'AircraftVelocity', 'FlightNumber', 'PassNumber')
    _required = (
        'Name', 'Type', 'Band', 'Bandwidth', 'CenterFrequency', 'Polarization', 'Range',
        'DepressionAngle', 'Aimpoint', 'AircraftHeading', 'AircraftTrackAngle',
        'Look', 'SquintAngle', 'AircraftLocation', 'AircraftVelocity')
    _numeric_format = {
        'Bandwidth': '0.17G', 'CenterFrequency': '0.17G', 'NearRange': '0.17G',
        'SlantRangeSwathWidth': '0.17G', 'Range': '0.17G', 'DepressionAngle': '0.17G',
        'LinearDynamicRange': '0.17G', 'AircraftHeading': '0.17G', 'AircraftTrackAngle': '0.17G', }
    # descriptors
    Name = StringDescriptor(
        'Name', _required,
        docstring='The name of the sensor')  # type: str
    SensorMfg = StringDescriptor(
        'SensorMfg', _required,
        docstring='The manufacturer of the sensor')  # type: Optional[str]
    OperatingAgency = StringDescriptor(
        'OperatingAgency', _required,
        docstring='The agency or company that operates the sensor')  # type: Optional[str]
    Type = StringDescriptor(
        'Type', _required,
        docstring='The type of sensor (i.e SAR or EO)')  # type: str
    Mode = StringDescriptor(
        'Mode', _required,
        docstring='Sensor operating mode')  # type: Optional[str]
    Band = StringDescriptor(
        'Band', _required,
        docstring='designation of the sensor frequency band')  # type: str
    Bandwidth = FloatDescriptor(
        'Bandwidth', _required,
        docstring='Radio Frequency bandwidth of the sensor system in GHz')  # type: float
    CenterFrequency = FloatDescriptor(
        'CenterFrequency', _required,
        docstring='Center operating frequency of the sensor system in GHz')  # type: float
    NearRange = FloatDescriptor(
        'NearRange', _required,
        docstring='The slant range distance measured from the sensor to the '
                  'near range of the image')  # type: Optional[float]
    SlantRangeSwathWidth = FloatDescriptor(
        'SlantRangeSwathWidth', _required,
        docstring='The width of the image as measured in the slant range'
    )  # type: Optional[float]
    Polarization = StringDescriptor(
        'Polarization', _required,
        docstring='The polarization of the transmitted/received signals')  # type: str
    Range = FloatDescriptor(
        'Range', _required,
        docstring='Measured slant range between the sensor aperture '
                  'and the scene center')  # type: float
    DepressionAngle = FloatDescriptor(
        'DepressionAngle', _required,
        docstring='Measured depression angle between the sensor line-of-sight '
                  'and the local horizontal reference plane')  # type: float
    LinearDynamicRange = FloatDescriptor(
        'LinearDynamicRange', _required,
        docstring="The span of the signal amplitudes (or power levels) over "
                  "which the system's response is linear. Typically the ratio "
                  "of the largest input signal that causes a 1 db compression "
                  "in receiver dynamic gain and the minimum signal defined by "
                  "receiver noise.")  # type: Optional[float]
    BeamWidth = SerializableDescriptor(
        'BeamWidth', BeamWidthType, _required,
        docstring='The width of the radar beam at its half power'
    )  # type: Optional[BeamWidthType]
    Aimpoint = SerializableDescriptor(
        'Aimpoint', LatLonEleType, _required,
        docstring='The sensor aim point')  # type: LatLonEleType
    AircraftHeading = FloatDescriptor(
        'AircraftHeading', _required,
        docstring='Aircraft heading relative to True North, in degrees'
    )  # type: float
    AircraftTrackAngle = FloatDescriptor(
        'AircraftTrackAngle', _required,
        docstring='The bearing from the aircraft position at the first pulse '
                  'to the aircraft position at the last')  # type: float
    Look = StringEnumDescriptor(
        'Look', {'Left', 'Right', 'Nadir'}, _required,
        docstring='Direction of the sensor look angle relative to aircraft '
                  'motion')  # type: str
    SquintAngle = SerializableDescriptor(
        'SquintAngle', SquintAngleType, _required,
        docstring='Measured angle between the sensor line-of-sight and the '
                  'lateral axis of the aircraft')  # type: SquintAngleType
    AircraftLocation = SerializableDescriptor(
        'AircraftLocation', AircraftLocationType, _required,
        docstring='The aircraft location (at scene center COA time?)')  # type: AircraftLocationType
    AircraftVelocity = SerializableDescriptor(
        'AircraftVelocity', XYZType, _required,
        docstring='Aircraft velocity in ECEF coordinates (at scene center COA time?)')  # type: XYZType
    FlightNumber = StringDescriptor(
        'FlightNumber', _required,
        docstring='The aircraft flight number')  # type: Optional[str]
    PassNumber = StringDescriptor(
        'PassNumber', _required,
        docstring='The aircraft pass number')  # type: Optional[str]

    def __init__(self, Name=None, SensorMfg=None, OperatingAgency=None,
                 Type=None, Mode=None, Band=None, Bandwidth=None,
                 CenterFrequency=None, NearRange=None, SlantRangeSwathWidth=None,
                 Polarization=None, Range=None, DepressionAngle=None,
                 LinearDynamicRange=None, BeamWidth=None, Aimpoint=None,
                 AircraftHeading=None, AircraftTrackAngle=None,
                 Look=None, SquintAngle=None,
                 AircraftLocation=None, AircraftVelocity=None,
                 FlightNumber=None, PassNumber=None, **kwargs):
        """
        Parameters
        ----------
        Name : None|str
        SensorMfg : None|str
        OperatingAgency : None|str
        Type : str
        Mode : None|str
        Band : None|str
        Bandwidth : None|float
        CenterFrequency : None|float
        NearRange : None|float
        SlantRangeSwathWidth : None|float
        Polarization : None|str
        Range : float
        DepressionAngle : float
        LinearDynamicRange : None|float
        BeamWidth : BeamWidthType
        Aimpoint : LatLonEleType|numpy.ndarray|list|tuple
        AircraftHeading : None|float
        AircraftTrackAngle : None|float
        Look : str
        SquintAngle : SquintAngleType
        AircraftLocation : AircraftLocationType|numpy.ndarray|list|tuple
        AircraftVelocity : XYZType|numpy.ndarray|list|tuple
        FlightNumber : None|int
        PassNumber : None|int

        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Name = Name
        self.SensorMfg = SensorMfg
        self.OperatingAgency = OperatingAgency
        self.Type = Type
        self.Mode = Mode
        self.Band = Band
        self.Bandwidth = Bandwidth
        self.CenterFrequency = CenterFrequency
        self.NearRange = NearRange
        self.SlantRangeSwathWidth = SlantRangeSwathWidth
        self.Polarization = Polarization
        self.Range = Range
        self.DepressionAngle = DepressionAngle
        self.LinearDynamicRange = LinearDynamicRange
        self.BeamWidth = BeamWidth
        self.Aimpoint = Aimpoint
        self.AircraftHeading = AircraftHeading
        self.AircraftTrackAngle = AircraftTrackAngle
        self.Look = Look
        self.SquintAngle = SquintAngle
        self.AircraftLocation = AircraftLocation
        self.AircraftVelocity = AircraftVelocity
        self.FlightNumber = FlightNumber
        self.PassNumber = PassNumber
        super(SensorInfoType, self).__init__(**kwargs)

    @classmethod
    def from_sicd(cls, sicd):
        """
        Construct the sensor info from a sicd structure

        Parameters
        ----------
        sicd : SICDType

        Returns
        -------
        SensorInfoType
        """

        transmit_freq_proc = sicd.ImageFormation.TxFrequencyProc
        center_freq = transmit_freq_proc.center_frequency*1e-9
        bandwidth = transmit_freq_proc.bandwidth*1e-9
        polarization = sicd.ImageFormation.get_polarization().replace(':', '')
        look = 'Left' if sicd.SCPCOA.SideOfTrack == 'L' else 'Right'
        arp_pos_llh = ecf_to_geodetic(sicd.SCPCOA.ARPPos.get_array())

        # calculate heading
        heading_ned = ecf_to_ned(sicd.SCPCOA.ARPVel.get_array(), sicd.SCPCOA.ARPPos.get_array(), absolute_coords=False)
        heading = numpy.rad2deg(numpy.arctan2(heading_ned[1], heading_ned[0]))
        # calculate track angle
        first_pos_ecf = sicd.Position.ARPPoly(0)
        last_pos_ecf = sicd.Position.ARPPoly(sicd.Timeline.CollectDuration)
        diff_ned = ecf_to_ned(last_pos_ecf - first_pos_ecf, sicd.SCPCOA.ARPPos.get_array(), absolute_coords=False)
        track_angle = numpy.rad2deg(numpy.arctan2(diff_ned[1], diff_ned[0]))

        return SensorInfoType(
            Name=sicd.CollectionInfo.CollectorName,
            Type='SAR',
            Mode=sicd.CollectionInfo.RadarMode.ModeType,
            Band=sicd.ImageFormation.get_transmit_band_name(),
            Bandwidth=bandwidth,
            CenterFrequency=center_freq,
            Polarization=polarization,
            Range=sicd.SCPCOA.SlantRange,
            DepressionAngle=sicd.SCPCOA.GrazeAng,
            Aimpoint=sicd.GeoData.SCP.LLH.get_array(),
            AircraftHeading=heading,
            AircraftTrackAngle=track_angle,
            Look=look,
            SquintAngle=SquintAngleType(
                SlantPlane=sicd.SCPCOA.DopplerConeAng,
                GroundPlane=sicd.SCPCOA.Squint),
            AircraftLocation=arp_pos_llh,
            AircraftVelocity=sicd.SCPCOA.ARPVel.get_array())




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class SECREC(TREElement):
    def __init__(self, value):
        super(SECREC, self).__init__()
        self.add_field('LOCID', 'b', 1, value)
        self.add_field('SECLEN', 'b', 1, value)
        self.add_field('PHYSIDX', 'b', 1, value)


class RPFIMGType(TREElement):
    def __init__(self, value):
        super(RPFIMGType, self).__init__()
        self.add_field('LOCLEN', 'b', 1, value)
        self.add_field('CLTOFF', 'b', 1, value)
        self.add_field('SECRECS', 'b', 1, value)
        self.add_field('RECLEN', 'b', 1, value)
        self.add_field('AGGLEN', 'b', 1, value)
        self.add_loop('SECRECList', int(self.SECRECS), SECREC, value)
        self.add_field('UNKNOWN', 'b', len(value) - self._bytes_length, value)


class RPFIMG(TREExtension):
    _tag_value = 'RPFIMG'
    _data_type = RPFIMGType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class RSMGIAType(TREElement):
    def __init__(self, value):
        super(RSMGIAType, self).__init__()
        self.add_field('IID', 's', 80, value)
        self.add_field('EDITION', 's', 40, value)
        self.add_field('GR0', 's', 21, value)
        self.add_field('GRX', 's', 21, value)
        self.add_field('GRY', 's', 21, value)
        self.add_field('GRZ', 's', 21, value)
        self.add_field('GRXX', 's', 21, value)
        self.add_field('GRXY', 's', 21, value)
        self.add_field('GRXZ', 's', 21, value)
        self.add_field('GRYY', 's', 21, value)
        self.add_field('GRYZ', 's', 21, value)
        self.add_field('GRZZ', 's', 21, value)
        self.add_field('GC0', 's', 21, value)
        self.add_field('GCX', 's', 21, value)
        self.add_field('GCY', 's', 21, value)
        self.add_field('GCZ', 's', 21, value)
        self.add_field('GCXX', 's', 21, value)
        self.add_field('GCXY', 's', 21, value)
        self.add_field('GCXZ', 's', 21, value)
        self.add_field('GCYY', 's', 21, value)
        self.add_field('GCYZ', 's', 21, value)
        self.add_field('GCZZ', 's', 21, value)
        self.add_field('GRNIS', 's', 3, value)
        self.add_field('GCNIS', 's', 3, value)
        self.add_field('GTNIS', 's', 3, value)
        self.add_field('GRSSIZ', 's', 21, value)
        self.add_field('GCSSIZ', 's', 21, value)


class RSMGIA(TREExtension):
    _tag_value = 'RSMGIA'
    _data_type = RSMGIAType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class RNT(TREElement):
    def __init__(self, value):
        super(RNT, self).__init__()
        self.add_field('RNPCF', 's', 21, value)


class RDT(TREElement):
    def __init__(self, value):
        super(RDT, self).__init__()
        self.add_field('RDPCF', 's', 21, value)


class CNT(TREElement):
    def __init__(self, value):
        super(CNT, self).__init__()
        self.add_field('CNPCF', 's', 21, value)


class CDT(TREElement):
    def __init__(self, value):
        super(CDT, self).__init__()
        self.add_field('CDPCF', 's', 21, value)


class RSMPCAType(TREElement):
    def __init__(self, value):
        super(RSMPCAType, self).__init__()
        self.add_field('IID', 's', 80, value)
        self.add_field('EDITION', 's', 40, value)
        self.add_field('RSN', 's', 3, value)
        self.add_field('CSN', 's', 3, value)
        self.add_field('RFEP', 's', 21, value)
        self.add_field('CFEP', 's', 21, value)
        self.add_field('RNRMO', 's', 21, value)
        self.add_field('CNRMO', 's', 21, value)
        self.add_field('XNRMO', 's', 21, value)
        self.add_field('YNRMO', 's', 21, value)
        self.add_field('ZNRMO', 's', 21, value)
        self.add_field('RNRMSF', 's', 21, value)
        self.add_field('CNRMSF', 's', 21, value)
        self.add_field('XNRMSF', 's', 21, value)
        self.add_field('YNRMSF', 's', 21, value)
        self.add_field('ZNRMSF', 's', 21, value)
        self.add_field('RNPWRX', 's', 1, value)
        self.add_field('RNPWRY', 's', 1, value)
        self.add_field('RNPWRZ', 's', 1, value)
        self.add_field('RNTRMS', 'd', 3, value)
        self.add_loop('RNTs', self.RNTRMS, RNT, value)
        self.add_field('RDPWRX', 's', 1, value)
        self.add_field('RDPWRY', 's', 1, value)
        self.add_field('RDPWRZ', 's', 1, value)
        self.add_field('RDTRMS', 'd', 3, value)
        self.add_loop('RDTs', self.RDTRMS, RDT, value)
        self.add_field('CNPWRX', 's', 1, value)
        self.add_field('CNPWRY', 's', 1, value)
        self.add_field('CNPWRZ', 's', 1, value)
        self.add_field('CNTRMS', 'd', 3, value)
        self.add_loop('CNTs', self.CNTRMS, CNT, value)
        self.add_field('CDPWRX', 's', 1, value)
        self.add_field('CDPWRY', 's', 1, value)
        self.add_field('CDPWRZ', 's', 1, value)
        self.add_field('CDTRMS', 'd', 3, value)
        self.add_loop('CDTs', self.CDTRMS, CDT, value)


class RSMPCA(TREExtension):
    _tag_value = 'RSMPCA'
    _data_type = RSMPCAType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class MPDSRAType(TREElement):
    def __init__(self, value):
        super(MPDSRAType, self).__init__()
        self.add_field('BLKNO', 's', 2, value)
        self.add_field('CDIPR', 's', 2, value)
        self.add_field('NBLKW', 's', 2, value)
        self.add_field('NRBLK', 's', 5, value)
        self.add_field('NCBLK', 's', 5, value)
        self.add_field('ORPX', 's', 9, value)
        self.add_field('ORPY', 's', 9, value)
        self.add_field('ORPZ', 's', 9, value)
        self.add_field('ORPRO', 's', 5, value)
        self.add_field('ORPCO', 's', 5, value)
        self.add_field('FPNVX', 's', 7, value)
        self.add_field('FPNVY', 's', 7, value)
        self.add_field('FPNVZ', 's', 7, value)
        self.add_field('ARPTM', 's', 9, value)
        self.add_field('RESV1', 's', 14, value)
        self.add_field('ARPPN', 's', 9, value)
        self.add_field('ARPPE', 's', 9, value)
        self.add_field('ARPPD', 's', 9, value)
        self.add_field('ARPVN', 's', 9, value)
        self.add_field('ARPVE', 's', 9, value)
        self.add_field('ARPVD', 's', 9, value)
        self.add_field('ARPAN', 's', 8, value)
        self.add_field('ARPAE', 's', 8, value)
        self.add_field('ARPAD', 's', 8, value)
        self.add_field('RESV2', 's', 13, value)


class MPDSRA(TREExtension):
    _tag_value = 'MPDSRA'
    _data_type = MPDSRAType



"""
The ProductDisplayType definition for SIDD 1.0.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List
from xml.etree import ElementTree
from collections import OrderedDict

import numpy

from sarpy.io.product.sidd2_elements.base import DEFAULT_STRICT
from sarpy.io.product.sidd2_elements.blocks import LatLonArrayElementType
from sarpy.io.xml.base import Serializable, SerializableArray, ParametersCollection, \
    find_children
from sarpy.io.xml.descriptors import SerializableDescriptor, SerializableArrayDescriptor, \
    SerializableListDescriptor, StringDescriptor, StringListDescriptor, ParametersDescriptor


class GeographicInformationType(Serializable):
    """
    Geographic information.
    """

    _fields = ('CountryCodes', 'SecurityInfo', 'GeographicInfoExtensions')
    _required = ()
    _collections_tags = {
        'CountryCodes': {'array': False, 'child_tag': 'CountryCode'},
        'GeographicInfoExtensions': {'array': False, 'child_tag': 'GeographicInfoExtension'}}
    # descriptors
    CountryCodes = StringListDescriptor(
        'CountryCodes', _required, strict=DEFAULT_STRICT,
        docstring="List of country codes for region covered by the image.")  # type: List[str]
    SecurityInfo = StringDescriptor(
        'SecurityInfo', _required, strict=DEFAULT_STRICT,
        docstring='Specifies classification level or special handling designators '
                  'for this geographic region.')  # type: Union[None, str]
    GeographicInfoExtensions = ParametersDescriptor(
        'GeographicInfoExtensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Implementation specific geographic information.')  # type: ParametersCollection

    def __init__(self, CountryCodes=None, SecurityInfo=None, GeographicInfoExtensions=None, **kwargs):
        """

        Parameters
        ----------
        CountryCodes : None|List[str]
        SecurityInfo : None|str
        GeographicInfoExtensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CountryCodes = CountryCodes
        self.SecurityInfo = SecurityInfo
        self.GeographicInfoExtensions = GeographicInfoExtensions
        super(GeographicInformationType, self).__init__(**kwargs)


class TargetInformationType(Serializable):
    """
    Information about the target.
    """
    _fields = ('Identifiers', 'Footprint', 'TargetInformationExtensions')
    _required = ()
    _collections_tags = {
        'Identifiers': {'array': False, 'child_tag': 'Identifier'},
        'Footprint': {'array': True, 'child_tag': 'Vertex'},
        'TargetInformationExtensions': {'array': False, 'child_tag': 'TargetInformationExtension'}}
    # Descriptors
    Identifiers = ParametersDescriptor(
        'Identifiers', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Target may have one or more identifiers.  Examples: names, BE numbers, etc. Use '
                  'the "name" attribute to describe what this is.')  # type: ParametersCollection
    Footprint = SerializableArrayDescriptor(
        'Footprint', LatLonArrayElementType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Target footprint as defined by polygonal '
                  'shape.')  # type: Union[SerializableArray, List[LatLonArrayElementType]]
    TargetInformationExtensions = ParametersDescriptor(
        'TargetInformationExtensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Generic extension. Could be used to indicate type of target, '
                  'terrain, etc.')  # type: ParametersCollection

    def __init__(self, Identifiers=None, Footprint=None, TargetInformationExtensions=None, **kwargs):
        """

        Parameters
        ----------
        Identifiers : None|ParametersCollection|dict
        Footprint : None|List[LatLonArrayElementType]|numpy.ndarray|list|tuple
        TargetInformationExtensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifiers = Identifiers
        self.Footprint = Footprint
        self.TargetInformationExtensions = TargetInformationExtensions
        super(TargetInformationType, self).__init__(**kwargs)


class GeographicCoverageType(Serializable):
    """
    The geographic coverage area for the product.
    """

    _fields = ('GeoregionIdentifiers', 'Footprint', 'GeographicInfo')
    _required = ('Footprint', )
    _collections_tags = {
        'GeoregionIdentifiers': {'array': False, 'child_tag': 'GeoregionIdentifier'},
        'Footprint': {'array': True, 'child_tag': 'Vertex'},
        'SubRegions': {'array': False, 'child_tag': 'SubRegion'}}
    # Descriptors
    GeoregionIdentifiers = ParametersDescriptor(
        'GeoregionIdentifiers', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Target may have one or more identifiers.  Examples: names, BE numbers, etc. Use '
                  'the "name" attribute to describe what this is.')  # type: ParametersCollection
    Footprint = SerializableArrayDescriptor(
        'Footprint', LatLonArrayElementType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Estimated ground footprint of the '
                  'product.')  # type: Union[None, SerializableArray, List[LatLonArrayElementType]]
    GeographicInfo = SerializableDescriptor(
        'GeographicInfo', GeographicInformationType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, GeographicInformationType]

    def __init__(self, GeoregionIdentifiers=None, Footprint=None, SubRegions=None, GeographicInfo=None, **kwargs):
        """

        Parameters
        ----------
        GeoregionIdentifiers : None|ParametersCollection|dict
        Footprint : None|List[LatLonArrayElementType]|numpy.ndarray|list|tuple
        SubRegions : None|List[GeographicCoverageType]
        GeographicInfo : None|GeographicInformationType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.GeoregionIdentifiers = GeoregionIdentifiers
        self.Footprint = Footprint
        self.GeographicInfo = GeographicInfo

        self._SubRegions = []
        if SubRegions is None:
            pass
        elif isinstance(SubRegions, GeographicCoverageType):
            self.addSubRegion(SubRegions)
        elif isinstance(SubRegions, (list, tuple)):
            for el in SubRegions:
                self.addSubRegion(el)
        else:
            raise ValueError('SubRegions got unexpected type {}'.format(type(SubRegions)))
        super(GeographicCoverageType, self).__init__(**kwargs)

    @property
    def SubRegions(self):
        """
        List[GeographicCoverageType]: list of sub-regions.
        """

        return self._SubRegions

    def addSubRegion(self, value):
        """
        Add the given SubRegion to the SubRegions list.

        Parameters
        ----------
        value : GeographicCoverageType

        Returns
        -------
        None
        """

        if isinstance(value, ElementTree.Element):
            value = GeographicCoverageType.from_node(value, self._xml_ns, ns_key=self._xml_ns_key)
        elif isinstance(value, dict):
            value = GeographicCoverageType.from_dict(value)

        if isinstance(value, GeographicCoverageType):
            self._SubRegions.append(value)
        else:
            raise TypeError('Trying to set SubRegion element with unexpected type {}'.format(type(value)))

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = OrderedDict()
        kwargs['SubRegions'] = find_children(node, 'SubRegion', xml_ns, ns_key)
        return super(GeographicCoverageType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        node = super(GeographicCoverageType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity, strict=strict, exclude=exclude)
        # slap on the SubRegion children
        sub_key = self._child_xml_ns_key.get('SubRegions', ns_key)
        for entry in self._SubRegions:
            entry.to_node(doc, 'SubRegion', ns_key=sub_key, parent=node, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(GeographicCoverageType, self).to_dict(check_validity=check_validity, strict=strict, exclude=exclude)
        # slap on the SubRegion children
        if len(self.SubRegions) > 0:
            out['SubRegions'] = [
                entry.to_dict(check_validity=check_validity, strict=strict) for entry in self._SubRegions]
        return out


class GeographicAndTargetType(Serializable):
    """
    Container specifying the image coverage area in geographic coordinates, as
    well as optional data about the target of the collection/product.
    """

    _fields = ('GeographicCoverage', 'TargetInformations')
    _required = ('GeographicCoverage', )
    _collections_tags = {'TargetInformations': {'array': False, 'child_tag': 'TargetInformation'}}
    # Descriptors
    GeographicCoverage = SerializableDescriptor(
        'GeographicCoverage', GeographicCoverageType, _required, strict=DEFAULT_STRICT,
        docstring='Provides geographic coverage information.')  # type: GeographicCoverageType
    TargetInformations = SerializableListDescriptor(
        'TargetInformations', TargetInformationType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Provides target specific geographic '
                  'information.')  # type: Union[None, List[TargetInformationType]]

    def __init__(self, GeographicCoverage=None, TargetInformations=None, **kwargs):
        """

        Parameters
        ----------
        GeographicCoverage : GeographicCoverageType
        TargetInformations : None|List[TargetInformationType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.GeographicCoverage = GeographicCoverage
        self.TargetInformations = TargetInformations
        super(GeographicAndTargetType, self).__init__(**kwargs)



"""
Common functionality for creating the SIDD structure from a SICD structure and
OrthorectificationHelper.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import numpy

from sarpy.io.complex.utils import two_dim_poly_fit, get_im_physical_coords
from sarpy.processing.ortho_rectify import OrthorectificationHelper, ProjectionHelper, \
    PGProjection
# agnostic to versions 1 & 2
from sarpy.io.product.sidd2_elements.Measurement import PlaneProjectionType, ProductPlaneType
from sarpy.io.product.sidd2_elements.blocks import ReferencePointType, Poly2DType, XYZPolyType, \
    FilterType, FilterBankType, PredefinedFilterType, NewLookupTableType, PredefinedLookupType
# version 3 elements
from sarpy.io.product.sidd3_elements.SIDD import SIDDType as SIDDType3
from sarpy.io.product.sidd3_elements.Display import ProductDisplayType as ProductDisplayType3, \
    NonInteractiveProcessingType as NonInteractiveProcessingType3, \
    ProductGenerationOptionsType as ProductGenerationOptionsType3, RRDSType as RRDSType3, \
    InteractiveProcessingType as InteractiveProcessingType3, GeometricTransformType as GeometricTransformType3, \
    SharpnessEnhancementType as SharpnessEnhancementType3, DynamicRangeAdjustmentType as DynamicRangeAdjustmentType3, \
    ScalingType as ScalingType3, OrientationType as OrientationType3
from sarpy.io.product.sidd3_elements.GeoData import GeoDataType as GeoDataType3
from sarpy.io.product.sidd3_elements.Measurement import MeasurementType as MeasurementType3, \
    PlaneProjectionType as PlaneProjectionType3, ProductPlaneType as ProductPlaneType3
from sarpy.io.product.sidd3_elements.ExploitationFeatures import ExploitationFeaturesType as ExploitationFeaturesType3
from sarpy.io.product.sidd3_elements.ProductCreation import ProductCreationType as ProductCreationType3
from sarpy.io.product.sidd3_elements.blocks import ReferencePointType as ReferencePointType3, \
    Poly2DType as Poly2DType3, XYZPolyType as XYZPolyType3, FilterType as FilterType3, \
    FilterBankType as FilterBankType3, PredefinedFilterType as PredefinedFilterType3, \
    NewLookupTableType as NewLookupTableType3, PredefinedLookupType as PredefinedLookupType3
# version 2 elements
from sarpy.io.product.sidd2_elements.SIDD import SIDDType as SIDDType2
from sarpy.io.product.sidd2_elements.Display import ProductDisplayType as ProductDisplayType2, \
    NonInteractiveProcessingType as NonInteractiveProcessingType2, \
    ProductGenerationOptionsType as ProductGenerationOptionsType2, RRDSType as RRDSType2, \
    InteractiveProcessingType as InteractiveProcessingType2, GeometricTransformType as GeometricTransformType2, \
    SharpnessEnhancementType as SharpnessEnhancementType2, DynamicRangeAdjustmentType as DynamicRangeAdjustmentType2, \
    ScalingType as ScalingType2, OrientationType as OrientationType2
from sarpy.io.product.sidd2_elements.GeoData import GeoDataType as GeoDataType2
from sarpy.io.product.sidd2_elements.Measurement import MeasurementType as MeasurementType2
from sarpy.io.product.sidd2_elements.ExploitationFeatures import ExploitationFeaturesType as ExploitationFeaturesType2
from sarpy.io.product.sidd2_elements.ProductCreation import ProductCreationType as ProductCreationType2
# version 1 elements
from sarpy.io.product.sidd1_elements.SIDD import SIDDType as SIDDType1
from sarpy.io.product.sidd1_elements.Display import ProductDisplayType as ProductDisplayType1
from sarpy.io.product.sidd1_elements.GeographicAndTarget import GeographicAndTargetType as GeographicAndTargetType1, \
    GeographicCoverageType as GeographicCoverageType1, GeographicInformationType as GeographicInformationType1
from sarpy.io.product.sidd1_elements.Measurement import MeasurementType as MeasurementType1
from sarpy.io.product.sidd1_elements.ExploitationFeatures import ExploitationFeaturesType as ExploitationFeaturesType1
from sarpy.io.product.sidd1_elements.ProductCreation import ProductCreationType as ProductCreationType1

logger = logging.getLogger(__name__)

_proj_helper_text = 'Unhandled projection helper type `{}`'

# TODO: move this to processing for 1.3.0


def _fit_timecoa_poly(proj_helper, bounds):
    """
    Fit the TimeCOA in new pixel coordinates.

    Parameters
    ----------
    proj_helper : ProjectionHelper
    bounds : numpy.ndarray
        The orthorectification pixel bounds of the form `(min row, max row, min col, max col)`.

    Returns
    -------
    Poly2DType
    """

    # what is the order of the sicd timecoapoly?
    in_poly = proj_helper.sicd.Grid.TimeCOAPoly
    use_order = max(in_poly.order1, in_poly.order2)
    if use_order == 0:
        # this is a constant polynomial, must be a spotlight collect
        return Poly2DType(Coefs=in_poly.get_array())
    # create an ortho coordinate grid
    samples = use_order+10
    ortho_grid = numpy.zeros((samples, samples, 2), dtype=numpy.float64)
    ortho_grid[:, :, 1], ortho_grid[:, :, 0] = numpy.meshgrid(
        numpy.linspace(bounds[2], bounds[3], num=samples),
        numpy.linspace(bounds[0], bounds[1], num=samples))
    # map to pixel grid coordinates
    pixel_grid = proj_helper.ortho_to_pixel(ortho_grid)
    pixel_rows_m = get_im_physical_coords(
        pixel_grid[:, :, 0], proj_helper.sicd.Grid, proj_helper.sicd.ImageData, 'row')
    pixel_cols_m = get_im_physical_coords(
        pixel_grid[:, :, 1], proj_helper.sicd.Grid, proj_helper.sicd.ImageData, 'col')
    # evaluate the sicd timecoapoly
    timecoa_values = proj_helper.sicd.Grid.TimeCOAPoly(pixel_rows_m, pixel_cols_m)
    sidd_timecoa_coeffs, residuals, rank, sing_values = two_dim_poly_fit(
        pixel_rows_m, pixel_cols_m, timecoa_values,
        x_order=use_order, y_order=use_order, x_scale=1e-3, y_scale=1e-3, rcond=1e-40)
    logger.info(
        'The time_coa_fit details:\n\t'
        'root mean square residuals = {}\n\t'
        'rank = {}\n\t'
        'singular values = {}'.format(residuals, rank, sing_values))
    return Poly2DType(Coefs=sidd_timecoa_coeffs)


def _create_plane_projection(proj_helper, bounds):
    """
    Construct the PlaneProjection structure for both version 1 & 2.

    Parameters
    ----------
    proj_helper : PGProjection
    bounds : numpy.ndarray
        The orthorectification pixel bounds of the form `(min row, max row, min col, max col)`.

    Returns
    -------
    PlaneProjectionType
    """

    ref_pixels = proj_helper.reference_pixels
    return PlaneProjectionType(
        ReferencePoint=ReferencePointType(ECEF=proj_helper.reference_point,
                                          Point=(float(ref_pixels[0]-bounds[0]), float(ref_pixels[1]-bounds[2]))),
        SampleSpacing=(proj_helper.row_spacing, proj_helper.col_spacing),
        TimeCOAPoly=_fit_timecoa_poly(proj_helper, bounds),
        ProductPlane=ProductPlaneType(RowUnitVector=proj_helper.row_vector,
                                      ColUnitVector=proj_helper.col_vector))


def _fit_timecoa_poly_v3(proj_helper, bounds):
    """
    Fit the TimeCOA in new pixel coordinates.

    Parameters
    ----------
    proj_helper : ProjectionHelper
    bounds : numpy.ndarray
        The orthorectification pixel bounds of the form `(min row, max row, min col, max col)`.

    Returns
    -------
    Poly2DType3
    """

    # what is the order of the sicd timecoapoly?
    in_poly = proj_helper.sicd.Grid.TimeCOAPoly
    use_order = max(in_poly.order1, in_poly.order2)
    if use_order == 0:
        # this is a constant polynomial, must be a spotlight collect
        return Poly2DType3(Coefs=in_poly.get_array())
    # create an ortho coordinate grid
    samples = use_order+10
    ortho_grid = numpy.zeros((samples, samples, 2), dtype=numpy.float64)
    ortho_grid[:, :, 1], ortho_grid[:, :, 0] = numpy.meshgrid(
        numpy.linspace(bounds[2], bounds[3], num=samples),
        numpy.linspace(bounds[0], bounds[1], num=samples))
    # map to pixel grid coordinates
    pixel_grid = proj_helper.ortho_to_pixel(ortho_grid)
    pixel_rows_m = get_im_physical_coords(
        pixel_grid[:, :, 0], proj_helper.sicd.Grid, proj_helper.sicd.ImageData, 'row')
    pixel_cols_m = get_im_physical_coords(
        pixel_grid[:, :, 1], proj_helper.sicd.Grid, proj_helper.sicd.ImageData, 'col')
    # evaluate the sicd timecoapoly
    timecoa_values = proj_helper.sicd.Grid.TimeCOAPoly(pixel_rows_m, pixel_cols_m)
    # fit this at the ortho_grid coordinates
    sidd_timecoa_coeffs, residuals, rank, sing_values = two_dim_poly_fit(
        pixel_rows_m, pixel_cols_m, timecoa_values,
        x_order=use_order, y_order=use_order, x_scale=1e-3, y_scale=1e-3, rcond=1e-40)
    logger.info(
        'The time_coa_fit details:\n\t'
        'root mean square residuals = {}\n\t'
        'rank = {}\n\t'
        'singular values = {}'.format(residuals, rank, sing_values))
    return Poly2DType3(Coefs=sidd_timecoa_coeffs)


def _create_plane_projection_v3(proj_helper, bounds):
    """
    Construct the PlaneProjection structure for both version 1 & 2.

    Parameters
    ----------
    proj_helper : PGProjection
    bounds : numpy.ndarray
    The orthorectification pixel bounds of the form `(min row, max row, min col, max col)`.

    Returns
    -------
    PlaneProjectionType3
    """

    ref_pixels = proj_helper.reference_pixels
    return PlaneProjectionType3(
        ReferencePoint=ReferencePointType3(ECEF=proj_helper.reference_point,
                                           Point=(float(ref_pixels[0]-bounds[0]),
                                                  float(ref_pixels[1]-bounds[2]))),
        SampleSpacing=(proj_helper.row_spacing, proj_helper.col_spacing),
        TimeCOAPoly=_fit_timecoa_poly_v3(proj_helper, bounds),
        ProductPlane=ProductPlaneType3(RowUnitVector=proj_helper.row_vector,
                                       ColUnitVector=proj_helper.col_vector))


#########################
# Version 3 element creation

def create_sidd_structure_v3(ortho_helper, bounds, product_class, pixel_type):
    """
    Create a SIDD version 3.0 structure based on the orthorectification helper
    and pixel bounds.

    Parameters
    ----------
    ortho_helper : OrthorectificationHelper
    bounds : numpy.ndarray|list|tuple
        The orthorectification pixel bounds of the form `(min row, max row, min col, max col)`.
    product_class : str
        A descriptive name for the product class. Examples -
        :code:`Dynamic Image, Amplitude Change Detection, Coherent Change Detection`
    pixel_type : str
        Must be one of `MONO8I, MONO16I` or `RGB24I`.

    Returns
    -------
    SIDDType3
    """

    def _create_display_v3():
        if pixel_type in ('MONO8I', 'MONO16I'):
            bands = 1
        elif pixel_type == 'RGB24I':
            bands = 3
        else:
            raise ValueError('pixel_type must be one of MONO8I, MONO16I, RGB24I. Got {}'.format(pixel_type))

        return ProductDisplayType3(
            PixelType=pixel_type,
            NumBands=bands,
            NonInteractiveProcessing=[NonInteractiveProcessingType3(
                ProductGenerationOptions=ProductGenerationOptionsType3(
                    DataRemapping=NewLookupTableType3(
                        LUTName='DENSITY',
                        Predefined=PredefinedLookupType3(
                            DatabaseName='DENSITY'))),
                RRDS=RRDSType3(DownsamplingMethod='DECIMATE'),
                band=i+1) for i in range(bands)],
            InteractiveProcessing=[InteractiveProcessingType3(
                GeometricTransform=GeometricTransformType3(
                    Scaling=ScalingType3(
                        AntiAlias=FilterType3(
                            FilterName='AntiAlias',
                            FilterBank=FilterBankType3(
                                Predefined=PredefinedFilterType3(DatabaseName='BILINEAR')),
                            Operation='CONVOLUTION'),
                        Interpolation=FilterType3(
                            FilterName='Interpolation',
                            FilterBank=FilterBankType3(
                                Predefined=PredefinedFilterType3(DatabaseName='BILINEAR')),
                            Operation='CORRELATION')),
                    Orientation=OrientationType3(ShadowDirection='ARBITRARY')),
                SharpnessEnhancement=SharpnessEnhancementType3(
                    ModularTransferFunctionEnhancement=FilterType3(
                        FilterName='ModularTransferFunctionEnhancement',
                        FilterBank=FilterBankType3(
                            Predefined=PredefinedFilterType3(DatabaseName='BILINEAR')),
                        Operation='CONVOLUTION')),
                DynamicRangeAdjustment=DynamicRangeAdjustmentType3(
                    AlgorithmType='NONE',
                    BandStatsSource=1),
                band=i+1) for i in range(bands)])

    def _create_measurement_v3():
        proj_helper = ortho_helper.proj_helper
        rows = bounds[1] - bounds[0]
        cols = bounds[3] - bounds[2]
        if isinstance(proj_helper, PGProjection):
            # fit the time coa polynomial in ortho-pixel coordinates
            plane_projection = _create_plane_projection_v3(proj_helper, bounds)
            return MeasurementType3(PixelFootprint=(rows, cols),
                                    ValidData=((0, 0), (0, cols), (rows, cols), (rows, 0)),
                                    PlaneProjection=plane_projection,
                                    ARPPoly=XYZPolyType3(
                                        X=proj_helper.sicd.Position.ARPPoly.X.get_array(),
                                        Y=proj_helper.sicd.Position.ARPPoly.Y.get_array(),
                                        Z=proj_helper.sicd.Position.ARPPoly.Z.get_array()))
        else:
            return None

    def _create_exploitation_v3():
        proj_helper = ortho_helper.proj_helper
        if isinstance(proj_helper, PGProjection):
            return ExploitationFeaturesType3.from_sicd(
                proj_helper.sicd, proj_helper.row_vector, proj_helper.col_vector)
        else:
            raise ValueError(_proj_helper_text.format(type(proj_helper)))

    pixel_type = pixel_type.upper()
    # validate bounds and get pixel coordinates rectangle
    bounds, ortho_pixel_corners = ortho_helper.bounds_to_rectangle(bounds)
    # construct appropriate SIDD elements
    prod_create = ProductCreationType3.from_sicd(ortho_helper.proj_helper.sicd, product_class)
    prod_create.Classification.ISMCATCESVersion = '201903'
    prod_create.Classification.compliesWith = 'USGov'

    # Display requires more product specifics
    display = _create_display_v3()
    # GeoData
    llh_corners = ortho_helper.proj_helper.ortho_to_llh(ortho_pixel_corners)
    geo_data = GeoDataType3(ImageCorners=llh_corners[:, :2], ValidData=llh_corners[:, :2])
    # Measurement
    measurement = _create_measurement_v3()
    # ExploitationFeatures
    exploit_feats = _create_exploitation_v3()
    return SIDDType3(ProductCreation=prod_create,
                     GeoData=geo_data,
                     Display=display,
                     Measurement=measurement,
                     ExploitationFeatures=exploit_feats)


#########################
# Version 2 element creation

def create_sidd_structure_v2(ortho_helper, bounds, product_class, pixel_type):
    """
    Create a SIDD version 2.0 structure based on the orthorectification helper
    and pixel bounds.

    Parameters
    ----------
    ortho_helper : OrthorectificationHelper
    bounds : numpy.ndarray|list|tuple
        The orthorectification pixel bounds of the form `(min row, max row, min col, max col)`.
    product_class : str
        A descriptive name for the product class. Examples -
        :code:`Dynamic Image, Amplitude Change Detection, Coherent Change Detection`
    pixel_type : str
        Must be one of `MONO8I, MONO16I` or `RGB24I`.

    Returns
    -------
    SIDDType2
    """

    def _create_display_v2():
        if pixel_type in ('MONO8I', 'MONO16I'):
            bands = 1
        elif pixel_type == 'RGB24I':
            bands = 3
        else:
            raise ValueError('pixel_type must be one of MONO8I, MONO16I, RGB24I. Got {}'.format(pixel_type))

        return ProductDisplayType2(
            PixelType=pixel_type,
            NumBands=bands,
            NonInteractiveProcessing=[NonInteractiveProcessingType2(
                ProductGenerationOptions=ProductGenerationOptionsType2(
                    DataRemapping=NewLookupTableType(
                        LUTName='DENSITY',
                        Predefined=PredefinedLookupType(
                            DatabaseName='DENSITY'))),
                RRDS=RRDSType2(DownsamplingMethod='DECIMATE'),
                band=i+1) for i in range(bands)],
            InteractiveProcessing=[InteractiveProcessingType2(
                GeometricTransform=GeometricTransformType2(
                    Scaling=ScalingType2(
                        AntiAlias=FilterType(
                            FilterName='AntiAlias',
                            FilterBank=FilterBankType(
                                Predefined=PredefinedFilterType(DatabaseName='BILINEAR')),
                            Operation='CONVOLUTION'),
                        Interpolation=FilterType(
                            FilterName='Interpolation',
                            FilterBank=FilterBankType(
                                Predefined=PredefinedFilterType(DatabaseName='BILINEAR')),
                            Operation='CORRELATION')),
                    Orientation=OrientationType2(ShadowDirection='ARBITRARY')),
                SharpnessEnhancement=SharpnessEnhancementType2(
                    ModularTransferFunctionEnhancement=FilterType(
                        FilterName='ModularTransferFunctionEnhancement',
                        FilterBank=FilterBankType(
                            Predefined=PredefinedFilterType(DatabaseName='BILINEAR')),
                        Operation='CONVOLUTION')),
                DynamicRangeAdjustment=DynamicRangeAdjustmentType2(
                    AlgorithmType='NONE',
                    BandStatsSource=1),
                band=i+1) for i in range(bands)])

    def _create_measurement_v2():
        proj_helper = ortho_helper.proj_helper
        rows = bounds[1] - bounds[0]
        cols = bounds[3] - bounds[2]
        if isinstance(proj_helper, PGProjection):
            # fit the time coa polynomial in ortho-pixel coordinates
            plane_projection = _create_plane_projection(proj_helper, bounds)
            return MeasurementType2(PixelFootprint=(rows, cols),
                                    ValidData=((0, 0), (0, cols), (rows, cols), (rows, 0)),
                                    PlaneProjection=plane_projection,
                                    ARPPoly=XYZPolyType(
                                        X=proj_helper.sicd.Position.ARPPoly.X.get_array(),
                                        Y=proj_helper.sicd.Position.ARPPoly.Y.get_array(),
                                        Z=proj_helper.sicd.Position.ARPPoly.Z.get_array()))
        else:
            return None

    def _create_exploitation_v2():
        proj_helper = ortho_helper.proj_helper
        if isinstance(proj_helper, PGProjection):
            return ExploitationFeaturesType2.from_sicd(
                proj_helper.sicd, proj_helper.row_vector, proj_helper.col_vector)
        else:
            raise ValueError(_proj_helper_text.format(type(proj_helper)))

    pixel_type = pixel_type.upper()
    # validate bounds and get pixel coordinates rectangle
    bounds, ortho_pixel_corners = ortho_helper.bounds_to_rectangle(bounds)
    # construct appropriate SIDD elements
    prod_create = ProductCreationType2.from_sicd(ortho_helper.proj_helper.sicd, product_class)
    prod_create.Classification.ISMCATCESVersion = '201903'
    prod_create.Classification.compliesWith = 'USGov'

    # Display requires more product specifics
    display = _create_display_v2()
    # GeoData
    llh_corners = ortho_helper.proj_helper.ortho_to_llh(ortho_pixel_corners)
    geo_data = GeoDataType2(ImageCorners=llh_corners[:, :2], ValidData=llh_corners[:, :2])
    # Measurement
    measurement = _create_measurement_v2()
    # ExploitationFeatures
    exploit_feats = _create_exploitation_v2()
    return SIDDType2(ProductCreation=prod_create,
                     GeoData=geo_data,
                     Display=display,
                     Measurement=measurement,
                     ExploitationFeatures=exploit_feats)


##########################
# Version 1 element creation

def create_sidd_structure_v1(ortho_helper, bounds, product_class, pixel_type):
    """
    Create a SIDD version 1.0 structure based on the orthorectification helper
    and pixel bounds.

    Parameters
    ----------
    ortho_helper : OrthorectificationHelper
    bounds : numpy.ndarray|list|tuple
        The orthorectification pixel bounds of the form `(min row, max row, min col, max col)`.
    product_class : str
        A descriptive name for the product class. Examples -
        :code:`Dynamic Image, Amplitude Change Detection, Coherent Change Detection`
    pixel_type : str
        Must be one of `MONO8I, MONO16I` or `RGB24I`.

    Returns
    -------
    SIDDType1
    """

    def _create_display_v1():
        if pixel_type not in ('MONO8I', 'MONO16I', 'RGB24I'):
            raise ValueError(
                'pixel_type must be one of MONO8I, MONO16I, RGB24I. Got {}'.format(pixel_type))

        return ProductDisplayType1(PixelType=pixel_type)

    def _create_measurement_v1():
        proj_helper = ortho_helper.proj_helper
        if isinstance(proj_helper, PGProjection):
            # fit the time coa polynomial in ortho-pixel coordinates
            plane_projection = _create_plane_projection(proj_helper, bounds)
            return MeasurementType1(PixelFootprint=(bounds[1] - bounds[0], bounds[3] - bounds[2]),
                                    PlaneProjection=plane_projection,
                                    ARPPoly=XYZPolyType(
                                        X=proj_helper.sicd.Position.ARPPoly.X.get_array(),
                                        Y=proj_helper.sicd.Position.ARPPoly.Y.get_array(),
                                        Z=proj_helper.sicd.Position.ARPPoly.Z.get_array()))
        else:
            raise ValueError(_proj_helper_text.format(type(proj_helper)))

    def _create_exploitation_v1():
        proj_helper = ortho_helper.proj_helper
        if isinstance(proj_helper, PGProjection):
            return ExploitationFeaturesType1.from_sicd(
                proj_helper.sicd, proj_helper.row_vector, proj_helper.col_vector)
        else:
            raise ValueError(_proj_helper_text.format(type(proj_helper)))

    pixel_type = pixel_type.upper()
    # validate bounds and get pixel coordinates rectangle
    bounds, ortho_pixel_corners = ortho_helper.bounds_to_rectangle(bounds)
    # construct appropriate SIDD elements
    prod_create = ProductCreationType1.from_sicd(ortho_helper.proj_helper.sicd, product_class)

    # Display requires more product specifics
    display = _create_display_v1()
    # GeographicAndTarget
    llh_corners = ortho_helper.proj_helper.ortho_to_llh(ortho_pixel_corners)
    geographic = GeographicAndTargetType1(
        GeographicCoverage=GeographicCoverageType1(Footprint=llh_corners[:, :2],
                                                   GeographicInfo=GeographicInformationType1()),)
    # Measurement
    measurement = _create_measurement_v1()
    # ExploitationFeatures
    exploit_feats = _create_exploitation_v1()

    return SIDDType1(ProductCreation=prod_create,
                     Display=display,
                     GeographicAndTarget=geographic,
                     Measurement=measurement,
                     ExploitationFeatures=exploit_feats)


##########################
# Switchable version SIDD structure

def create_sidd_structure(ortho_helper, bounds, product_class, pixel_type, version=3):
    """
    Create a SIDD structure, with version specified, based on the orthorectification
    helper and pixel bounds.

    Parameters
    ----------
    ortho_helper : OrthorectificationHelper
    bounds : numpy.ndarray|list|tuple
        The orthorectification pixel bounds of the form `(min row, max row, min col, max col)`.
    product_class : str
        A descriptive name for the product class. Examples -
        :code:`Dynamic Image, Amplitude Change Detection, Coherent Change Detection`
    pixel_type : str
        Must be one of `MONO8I, MONO16I` or `RGB24I`.
    version : int
        The SIDD version, must be either 1, 2, or 3.

    Returns
    -------
    SIDDType1|SIDDType2|SIDDType3
    """

    if version not in [1, 2, 3]:
        raise ValueError('version must be 1, 2, or 3. Got {}'.format(version))

    if version == 1:
        return create_sidd_structure_v1(ortho_helper, bounds, product_class, pixel_type)
    elif version == 2:
        return create_sidd_structure_v2(ortho_helper, bounds, product_class, pixel_type)
    else:
        return create_sidd_structure_v3(ortho_helper, bounds, product_class, pixel_type)



"""
The GeoData definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from collections import OrderedDict
from xml.etree import ElementTree
from typing import List, Union, Dict, Optional

import numpy

from sarpy.io.xml.base import Serializable, SerializableArray, ParametersCollection, \
    find_children, parse_serializable
from sarpy.io.xml.descriptors import StringDescriptor, StringEnumDescriptor, \
    SerializableDescriptor, ParametersDescriptor, SerializableArrayDescriptor

from .base import DEFAULT_STRICT, SerializableCPArrayDescriptor, SerializableCPArray
from .blocks import XYZType, LatLonRestrictionType, LatLonHAERestrictionType, \
    LatLonCornerStringType, LatLonArrayElementType

from sarpy.geometry.geocoords import geodetic_to_ecf, ecf_to_geodetic


class GeoInfoType(Serializable):
    """
    A geographic feature.
    """

    _fields = ('name', 'Descriptions', 'Point', 'Line', 'Polygon')
    _required = ('name', )
    _set_as_attribute = ('name', )
    _choice = ({'required': False, 'collection': ('Point', 'Line', 'Polygon')}, )
    _collections_tags = {
        'Descriptions': {'array': False, 'child_tag': 'Desc'},
        'Line': {'array': True, 'child_tag': 'Endpoint'},
        'Polygon': {'array': True, 'child_tag': 'Vertex'}, }
    # descriptors
    name = StringDescriptor(
        'name', _required, strict=DEFAULT_STRICT,
        docstring='The name.')  # type: str
    Descriptions = ParametersDescriptor(
        'Descriptions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Descriptions of the geographic feature.')  # type: ParametersCollection
    Point = SerializableDescriptor(
        'Point', LatLonRestrictionType, _required, strict=DEFAULT_STRICT,
        docstring='A geographic point with WGS-84 coordinates.')  # type: LatLonRestrictionType
    Line = SerializableArrayDescriptor(
        'Line', LatLonArrayElementType, _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=2,
        docstring='A geographic line (array) with WGS-84 coordinates.'
    )  # type: Union[SerializableArray, List[LatLonArrayElementType]]
    Polygon = SerializableArrayDescriptor(
        'Polygon', LatLonArrayElementType, _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=3,
        docstring='A geographic polygon (array) with WGS-84 coordinates.'
    )  # type: Union[SerializableArray, List[LatLonArrayElementType]]

    def __init__(
            self,
            name: str = None,
            Descriptions: Union[None, ParametersCollection, Dict] = None,
            Point=None,
            Line=None,
            Polygon=None,
            GeoInfos=None,
            **kwargs):
        """

        Parameters
        ----------
        name : str
        Descriptions : None|ParametersCollection|dict
        Point : None|LatLonRestrictionType|numpy.ndarray|list|tuple
        Line : None|SerializableArray|List[LatLonArrayElementType]|numpy.ndarray|list|tuple
        Polygon : None|SerializableArray|List[LatLonArrayElementType]|numpy.ndarray|list|tuple
        GeoInfos : None|Sequence[GeoInfoType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.name = name
        self.Descriptions = Descriptions
        self.Point = Point
        self.Line = Line
        self.Polygon = Polygon

        self._GeoInfos = []
        if GeoInfos is None:
            pass
        elif isinstance(GeoInfos, GeoInfoType):
            self.addGeoInfo(GeoInfos)
        elif isinstance(GeoInfos, (list, tuple)):
            for el in GeoInfos:
                self.addGeoInfo(el)
        else:
            raise ValueError('GeoInfos got unexpected type {}'.format(type(GeoInfos)))
        super(GeoInfoType, self).__init__(**kwargs)

    @property
    def FeatureType(self) -> Optional[str]:
        """
        str: READ ONLY attribute. Identifies the feature type among. This is determined by
        returning the (first) attribute among `Point`, `Line`, `Polygon` which is populated.
        `None` will be returned if none of them are populated.
        """

        for attribute in self._choice[0]['collection']:
            if getattr(self, attribute) is not None:
                return attribute
        return None

    @property
    def GeoInfos(self):
        """
        List[GeoInfoType]: list of GeoInfos.
        """

        return self._GeoInfos

    def getGeoInfo(self, key: str):
        """
        Get GeoInfo(s) with name attribute == `key`.

        Parameters
        ----------
        key : str

        Returns
        -------
        List[GeoInfoType]
        """

        return [entry for entry in self._GeoInfos if entry.name == key]

    def addGeoInfo(self, value):
        """
        Add the given GeoInfo to the GeoInfos list.

        Parameters
        ----------
        value : GeoInfoType|Dict

        Returns
        -------
        None
        """

        if isinstance(value, ElementTree.Element):
            gi_key = self._child_xml_ns_key.get('GeoInfos', self._xml_ns_key)
            value = self.from_node(value, self._xml_ns, ns_key=gi_key)
        elif isinstance(value, dict):
            value = self.from_dict(value)

        if isinstance(value, GeoInfoType):
            self._GeoInfos.append(value)
        else:
            raise TypeError('Trying to set GeoInfo element with unexpected type {}'.format(type(value)))

    def _validate_features(self) -> bool:
        if self.Line is not None and self.Line.size < 2:
            self.log_validity_error('GeoInfo has a Line feature with {} points defined.'.format(self.Line.size))
            return False
        if self.Polygon is not None and self.Polygon.size < 3:
            self.log_validity_error('GeoInfo has a Polygon feature with {} points defined.'.format(self.Polygon.size))
            return False
        return True

    def _basic_validity_check(self) -> bool:
        condition = super(GeoInfoType, self)._basic_validity_check()
        return condition & self._validate_features()

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = OrderedDict()
        gi_key = cls._child_xml_ns_key.get('GeoInfos', ns_key)
        kwargs['GeoInfos'] = find_children(node, 'GeoInfo', xml_ns, gi_key)
        return super(GeoInfoType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        node = super(GeoInfoType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity, strict=strict, exclude=exclude)
        # slap on the GeoInfo children
        for entry in self._GeoInfos:
            entry.to_node(doc, tag, ns_key=ns_key, parent=node, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(GeoInfoType, self).to_dict(check_validity=check_validity, strict=strict, exclude=exclude)
        # slap on the GeoInfo children
        if len(self.GeoInfos) > 0:
            out['GeoInfos'] = [entry.to_dict(check_validity=check_validity, strict=strict) for entry in self._GeoInfos]
        return out


class SCPType(Serializable):
    """
    Scene Center Point (SCP) in full (global) image. This should be the precise location.
    Note that setting one of ECF or LLH will implicitly set the other to it's corresponding matched value.
    """

    _fields = ('ECF', 'LLH')
    _required = _fields
    _ECF = None
    _LLH = None

    def __init__(
            self,
            ECF: Union[None, XYZType, numpy.ndarray, tuple, list] = None,
            LLH: Union[None, LatLonHAERestrictionType, numpy.ndarray, tuple, list] = None,
            **kwargs):
        """
        To avoid the potential of inconsistent state, ECF and LLH are not simultaneously
        used. If ECF is provided, it is used to populate LLH. Otherwise, if LLH is provided,
        then it is used to populate ECF.

        Parameters
        ----------
        ECF : XYZType|numpy.ndarray|list|tuple
        LLH : LatLonHAERestrictionType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        if ECF is not None:
            self.ECF = ECF
        elif LLH is not None:
            self.LLH = LLH
        super(SCPType, self).__init__(**kwargs)

    @property
    def ECF(self) -> XYZType:
        """
        XYZType: The ECF coordinates.
        """

        return self._ECF

    @ECF.setter
    def ECF(self, value):
        if value is not None:
            self._ECF = parse_serializable(value, 'ECF', self, XYZType)
            self._LLH = LatLonHAERestrictionType.from_array(ecf_to_geodetic(self._ECF.get_array()))

    @property
    def LLH(self) -> LatLonHAERestrictionType:
        """
        LatLonHAERestrictionType: The WGS-84 coordinates.
        """

        return self._LLH

    @LLH.setter
    def LLH(self, value):
        if value is not None:
            self._LLH = parse_serializable(value, 'LLH', self, LatLonHAERestrictionType)
            self._ECF = XYZType.from_array(geodetic_to_ecf(self._LLH.get_array(order='LAT')))

    def get_image_center_abbreviation(self) -> str:
        """
        Gets the center coordinate abbreviation for the suggested name.

        Returns
        -------
        str
        """

        llh = self.LLH
        lat = int(numpy.round(llh.Lat))
        lon = int(numpy.round(llh.Lon))
        return '{0:02d}'.format(abs(lat)) + ('N' if lat >= 0 else 'S') + \
            '{0:03d}'.format(abs(lon)) + ('E' if lon >= 0 else 'W')


class GeoDataType(Serializable):
    """Container specifying the image coverage area in geographic coordinates."""
    _fields = ('EarthModel', 'SCP', 'ImageCorners', 'ValidData')
    _required = ('EarthModel', 'SCP', 'ImageCorners')
    _collections_tags = {
        'ValidData': {'array': True, 'child_tag': 'Vertex'},
        'ImageCorners': {'array': True, 'child_tag': 'ICP'},
    }
    # other class variables
    _EARTH_MODEL_VALUES = ('WGS_84', )
    # descriptors
    EarthModel = StringEnumDescriptor(
        'EarthModel', _EARTH_MODEL_VALUES, _required, strict=True, default_value='WGS_84',
        docstring='The Earth Model.')  # type: str
    SCP = SerializableDescriptor(
        'SCP', SCPType, _required, strict=DEFAULT_STRICT,
        docstring='The Scene Center Point *(SCP)* in full (global) image. This is the '
                  'precise location.')  # type: SCPType
    ImageCorners = SerializableCPArrayDescriptor(
        'ImageCorners', LatLonCornerStringType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The geographic image corner points array. Image corners points projected to the '
                  'ground/surface level. Points may be projected to the same height as the SCP if ground/surface '
                  'height data is not available. The corner positions are approximate geographic locations and '
                  'not intended for analytical use.')  # type: Union[SerializableCPArray, List[LatLonCornerStringType]]
    ValidData = SerializableArrayDescriptor(
        'ValidData', LatLonArrayElementType, _collections_tags, _required,
        strict=DEFAULT_STRICT, minimum_length=3,
        docstring='The full image array includes both valid data and some zero filled pixels.'
    )  # type: Union[SerializableArray, List[LatLonArrayElementType]]

    def __init__(
            self,
            EarthModel: str = 'WGS_84',
            SCP: SCPType = None,
            ImageCorners=None,
            ValidData=None,
            GeoInfos: List[GeoInfoType] = None,
            **kwargs):
        """

        Parameters
        ----------
        EarthModel : str
        SCP : SCPType
        ImageCorners : SerializableCPArray|List[LatLonCornerStringType]|numpy.ndarray|list|tuple
        ValidData : SerializableArray|List[LatLonArrayElementType]|numpy.ndarray|list|tuple
        GeoInfos : List[GeoInfoType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.EarthModel = EarthModel
        self.SCP = SCP
        self.ImageCorners = ImageCorners
        self.ValidData = ValidData

        self._GeoInfos = []
        if GeoInfos is None:
            pass
        elif isinstance(GeoInfos, GeoInfoType):
            self.setGeoInfo(GeoInfos)
        elif isinstance(GeoInfos, (list, tuple)):
            for el in GeoInfos:
                self.setGeoInfo(el)
        else:
            raise ValueError('GeoInfos got unexpected type {}'.format(type(GeoInfos)))
        super(GeoDataType, self).__init__(**kwargs)

    def derive(self):
        """
        Populates any potential derived data in GeoData. Is expected to be called by
        the `SICD` parent as part of a more extensive derived data effort.

        Returns
        -------
        None
        """

        pass

    @property
    def GeoInfos(self) -> List[GeoInfoType]:
        """
        List[GeoInfoType]: list of GeoInfos.
        """

        return self._GeoInfos

    def getGeoInfo(self, key: str) -> List[GeoInfoType]:
        """
        Get the GeoInfo(s) with name attribute == `key`

        Parameters
        ----------
        key : str

        Returns
        -------
        List[GeoInfoType]
        """

        return [entry for entry in self._GeoInfos if entry.name == key]

    def setGeoInfo(self, value: Union[GeoInfoType, Dict]):
        """
        Add the given GeoInfo to the GeoInfos list.

        Parameters
        ----------
        value : GeoInfoType

        Returns
        -------
        None
        """

        if isinstance(value, ElementTree.Element):
            gi_key = self._child_xml_ns_key.get('GeoInfos', self._xml_ns_key)
            value = GeoInfoType.from_node(value, self._xml_ns, ns_key=gi_key)
        elif isinstance(value, dict):
            value = GeoInfoType.from_dict(value)

        if isinstance(value, GeoInfoType):
            self._GeoInfos.append(value)
        else:
            raise TypeError('Trying to set GeoInfo element with unexpected type {}'.format(type(value)))

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        if kwargs is None:
            kwargs = OrderedDict()
        gi_key = cls._child_xml_ns_key.get('GeoInfos', ns_key)
        kwargs['GeoInfos'] = find_children(node, 'GeoInfo', xml_ns, gi_key)
        return super(GeoDataType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)

    def to_node(self, doc, tag, ns_key=None, parent=None, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        node = super(GeoDataType, self).to_node(
            doc, tag, ns_key=ns_key, parent=parent, check_validity=check_validity, strict=strict, exclude=exclude)
        # slap on the GeoInfo children
        for entry in self._GeoInfos:
            entry.to_node(doc, 'GeoInfo', ns_key=ns_key, parent=node, strict=strict)
        return node

    def to_dict(self, check_validity=False, strict=DEFAULT_STRICT, exclude=()):
        out = super(GeoDataType, self).to_dict(check_validity=check_validity, strict=strict, exclude=exclude)
        # slap on the GeoInfo children
        if len(self.GeoInfos) > 0:
            out['GeoInfos'] = [entry.to_dict(check_validity=check_validity, strict=strict) for entry in self._GeoInfos]
        return out

    def _basic_validity_check(self) -> bool:
        condition = super(GeoDataType, self)._basic_validity_check()
        return condition



"""
Definition for the main NGA modified RDE/AFRL labeling object
"""

__classification__ = "UNCLASSIFIED"
__authors__ = "Thomas McCullough"

from typing import Optional
import os

from sarpy.io.xml.base import Serializable, parse_xml_from_string, parse_xml_from_file
from sarpy.io.xml.descriptors import SerializableDescriptor, StringDescriptor

from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd import SICDReader
from sarpy.io.general.utils import calculate_md5

from .base import DEFAULT_STRICT
from .CollectionInfo import CollectionInfoType
from .SubCollectionInfo import SubCollectionInfoType
from .ImageInfo import ImageInfoType
from .SensorInfo import SensorInfoType
from .FiducialInfo import FiducialInfoType
from .ObjectInfo import ObjectInfoType


_AFRL_SPECIFICATION_NAMESPACE = 'urn:AFRL_RDE:1.0.0'


class ResearchType(Serializable):
    _fields = (
        'MetadataVersion', 'DetailCollectionInfo', 'DetailSubCollectionInfo',
        'DetailImageInfo', 'DetailSensorInfo', 'DetailFiducialInfo', 'DetailObjectInfo')
    _required = (
        'MetadataVersion', 'DetailCollectionInfo', 'DetailSubCollectionInfo',
        'DetailImageInfo', 'DetailSensorInfo', 'DetailFiducialInfo', 'DetailObjectInfo')
    # descriptors
    MetadataVersion = StringDescriptor(
        'MetadataVersion', _required,
        docstring='The version number')  # type: str
    DetailCollectionInfo = SerializableDescriptor(
        'DetailCollectionInfo', CollectionInfoType, _required,
        docstring='High level information about the data collection'
    )  # type: Optional[CollectionInfoType]
    DetailSubCollectionInfo = SerializableDescriptor(
        'DetailSubCollectionInfo', SubCollectionInfoType, _required,
        docstring='Information about sub-division of the overall data collection'
    )  # type: Optional[SubCollectionInfoType]
    DetailImageInfo = SerializableDescriptor(
        'DetailImageInfo', ImageInfoType, _required,
        docstring='Information about the referenced image'
    )  # type: Optional[ImageInfoType]
    DetailSensorInfo = SerializableDescriptor(
        'DetailSensorInfo', SensorInfoType, _required,
        docstring='Information about the sensor'
    )  # type: Optional[SensorInfoType]
    DetailFiducialInfo = SerializableDescriptor(
        'DetailFiducialInfo', FiducialInfoType, _required,
        docstring='Information about the ground-truthed fiducials'
    )  # type: Optional[FiducialInfoType]
    DetailObjectInfo = SerializableDescriptor(
        'DetailObjectInfo', ObjectInfoType, _required,
        docstring='Information about the ground-truthed objects'
    )  # type: Optional[ObjectInfoType]

    def __init__(self, MetadataVersion='Unknown', DetailCollectionInfo=None,
                 DetailSubCollectionInfo=None, DetailImageInfo=None,
                 DetailSensorInfo=None, DetailFiducialInfo=None,
                 DetailObjectInfo=None, **kwargs):
        """
        Parameters
        ----------
        MetadataVersion : str
        DetailCollectionInfo : CollectionInfoType
        DetailSubCollectionInfo : SubCollectionInfoType
        DetailImageInfo : ImageInfoType
        DetailSensorInfo : SensorInfo
        DetailFiducialInfo : FiducialInfoType
        DetailObjectInfo : ObjectInfoType
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.MetadataVersion = MetadataVersion
        self.DetailCollectionInfo = DetailCollectionInfo
        self.DetailSubCollectionInfo = DetailSubCollectionInfo
        self.DetailImageInfo = DetailImageInfo
        self.DetailSensorInfo = DetailSensorInfo
        self.DetailFiducialInfo = DetailFiducialInfo
        self.DetailObjectInfo = DetailObjectInfo
        super(ResearchType, self).__init__(**kwargs)

    def to_xml_bytes(self, urn=None, tag='RESEARCH', check_validity=False, strict=DEFAULT_STRICT):
        if urn is None:
            urn = _AFRL_SPECIFICATION_NAMESPACE
        return super(ResearchType, self).to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict)

    def to_xml_string(self, urn=None, tag='RESEARCH', check_validity=False, strict=DEFAULT_STRICT):
        return self.to_xml_bytes(urn=urn, tag=tag, check_validity=check_validity, strict=strict).decode('utf-8')

    def apply_sicd(self, sicd, base_file_name, layover_shift=False, populate_in_periphery=False, include_out_of_range=False,
                   padding_fraction=0.05, minimum_pad=0, md5_checksum=None):
        """
        Apply the given sicd to define all the relevant derived data, assuming
        that the starting point is physical ground truth populated, and image
        details and locations will be inferred. This modifies the structure in
        place.

        Parameters
        ----------
        sicd : SICDType
        base_file_name : str
        layover_shift : bool
        populate_in_periphery : bool
        include_out_of_range : bool
        padding_fraction : None|float
        minimum_pad : int|float
        md5_checksum : None|str
        """

        # assume that collection info and subcollection info are previously defined

        # define the image info
        if self.DetailImageInfo is not None:
            raise ValueError('Image Info is already defined')
        self.DetailImageInfo = ImageInfoType.from_sicd(
            sicd, base_file_name, md5_checksum=md5_checksum)

        # define sensor info
        if self.DetailSensorInfo is not None:
            raise ValueError('Sensor Info is already defined')
        self.DetailSensorInfo = SensorInfoType.from_sicd(sicd)

        if self.DetailFiducialInfo is None:
            self.DetailFiducialInfo = FiducialInfoType(
                NumberOfFiducialsInImage=0, NumberOfFiducialsInScene=0)
        else:
            self.DetailFiducialInfo.set_image_location_from_sicd(
                sicd,
                populate_in_periphery=populate_in_periphery,
                include_out_of_range=include_out_of_range)

        if self.DetailObjectInfo is None:
            self.DetailObjectInfo = ObjectInfoType(
                NumberOfObjectsInImage=0, NumberOfObjectsInScene=0)
        else:
            self.DetailObjectInfo.set_image_location_from_sicd(
                sicd,
                layover_shift=layover_shift,
                populate_in_periphery=populate_in_periphery,
                include_out_of_range=include_out_of_range,
                padding_fraction=padding_fraction,
                minimum_pad=minimum_pad)

    def apply_sicd_reader(
            self, sicd_reader, layover_shift=False, populate_in_periphery=False, include_out_of_range=False,
            padding_fraction=0.05, minimum_pad=0, populate_md5=True):
        """
        Apply the given sicd to define all the relevant derived data, assuming
        that the starting point is physical ground truth populated, and image
        details and locations will be inferred. This modifies the structure in
        place.

        Parameters
        ----------
        sicd_reader : SICDReader
        layover_shift : bool
        populate_in_periphery : bool
        include_out_of_range : bool
        padding_fraction : None|float
        minimum_pad : int|float
        populate_md5 : bool
        """

        md5_checksum = None if (sicd_reader.file_name is None or not populate_md5) \
            else calculate_md5(sicd_reader.file_name)

        base_file = os.path.split(sicd_reader.file_name)[1]
        self.apply_sicd(
            sicd_reader.sicd_meta,
            base_file,
            layover_shift=layover_shift,
            populate_in_periphery=populate_in_periphery,
            include_out_of_range=include_out_of_range,
            padding_fraction=padding_fraction,
            minimum_pad=minimum_pad,
            md5_checksum=md5_checksum)

    @classmethod
    def from_xml_file(cls, file_path):
        """
        Construct the research object from an xml file path.

        Parameters
        ----------
        file_path : str

        Returns
        -------
        ResearchType
        """

        root_node, xml_ns = parse_xml_from_file(file_path)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)

    @classmethod
    def from_xml_string(cls, xml_string):
        """
        Construct the research object from an xml string.

        Parameters
        ----------
        xml_string : str|bytes

        Returns
        -------
        ResearchType
        """

        root_node, xml_ns = parse_xml_from_string(xml_string)
        ns_key = 'default' if 'default' in xml_ns else None
        return cls.from_node(root_node, xml_ns=xml_ns, ns_key=ns_key)
